from typing import NamedTuple
import re
from celline.config import Config

class TemplateManager:
    @staticmethod
    def replace_from_file(file_name: str, structure: NamedTuple, replaced_path: str):
        with open(
            f"{Config.EXEC_ROOT}/celline/template/content/{file_name}",
            "r",
            encoding="utf-8",
        ) as file:
            data = file.read()
        # shebang = f"#!/usr/bin/env {os.path.basename(os.environ.get('SHELL', '/bin/sh'))}\n"
        # data = shebang + data
        for field in structure._fields:
            data = re.sub(f"%{field}/", getattr(structure, field), data)

        with open(replaced_path, "w", encoding="utf-8") as file:
            file.write(data)



import datetime
import glob
import os
from typing import Final, List

from celline.config import Config


class Path:
    project_id: Final[str]
    sample_id: Final[str]

    def __init__(self, project_id: str, sample_id: str) -> None:
        self.project_id = project_id
        self.sample_id = sample_id

    @property
    def resources(self):
        return f"{Config.PROJ_ROOT}/resources"

    @property
    def data(self):
        return f"{Config.PROJ_ROOT}/data"

    @property
    def resources_sample(self):
        return f"{self.resources}/{self.project_id}/{self.sample_id}"

    @property
    def resources_sample_src(self):
        return f"{self.resources_sample}/src"

    @property
    def resources_sample_counted(self):
        return f"{self.resources_sample}/counted"

    @property
    def resources_sample_raw(self):
        return f"{self.resources_sample}/raw"

    @property
    def resources_sample_raw_fastqs(self):
        return f"{self.resources_sample_raw}/fastqs"

    @property
    def resources_sample_log(self):
        return f"{self.resources_sample}/logs"

    def resources_log_file(self, prefix: str):
        return f"{self.resources_sample_log}/{prefix}_{datetime.datetime.now().strftime('%Y%m%d_%H:%M:%S')}.log"

    @property
    def data_sample(self):
        return f"{self.data}/{self.project_id}/{self.sample_id}"
    @property
    def data_sample_predicted_celltype(self):
        return f"{self.data}/{self.project_id}/{self.sample_id}/celltype_predicted.tsv"

    @property
    def data_sample_log(self):
        return f"{self.data}/{self.project_id}/{self.sample_id}/logs"

    @property
    def data_sample_src(self):
        return f"{self.data}/{self.project_id}/{self.sample_id}/src"

    def data_log_file(self, prefix: str):
        return f"{self.data_sample_log}/{prefix}_{datetime.datetime.now().strftime('%Y%m%d_%H:%M:%S')}.log"

    @property
    def is_downloaded(self):
        patterns = [
            f"{self.sample_id}_S1_L00*_{suffix}_001.fastq.gz"
            for suffix in ["R1", "R2", "I1", "I2"]
        ]

        all_files: List[bool] = []
        for file in glob.glob(f"{self.resources_sample_raw_fastqs}/*"):
            total_files = 0
            for pattern in patterns:
                files = glob.glob(os.path.join(file, pattern))
                total_files += len(files)
            if 2 <= total_files:
                all_files.append(True)
            else:
                all_files.append(False)
        return len(all_files) > 0 and all(all_files)

    @property
    def is_counted(self):
        """Already counted?"""
        counted_file = (
            f"{self.resources_sample_counted}/outs/filtered_feature_bc_matrix.h5"
        )
        return os.path.isfile(counted_file)

    @property
    def is_doublet_predicted(self) -> bool:
        return os.path.isfile(f"{self.data_sample}/doublet_info.tsv")

    @property
    def is_preprocessed(self) -> bool:
        return os.path.isfile(
            f"{self.data_sample}/doublet_info.tsv"
        ) and os.path.isfile(f"{self.data_sample}/qc_matrix.tsv")

    @property
    def is_predicted_celltype(self):
        return os.path.isfile(f"{self.data_sample}/celltype_predicted.tsv")

    def prepare(self):
        if not os.path.isdir(self.resources_sample_raw):
            os.makedirs(self.resources_sample_raw, exist_ok=True)
        if not os.path.isdir(self.resources_sample_log):
            os.makedirs(self.resources_sample_log, exist_ok=True)
        if not os.path.isdir(self.resources_sample_src):
            os.makedirs(self.resources_sample_src, exist_ok=True)
        if not os.path.isdir(self.data_sample_log):
            os.makedirs(self.data_sample_log, exist_ok=True)
        if not os.path.isdir(self.data_sample_src):
            os.makedirs(self.data_sample_src, exist_ok=True)



"""
Main CLI entry point for celline.
"""

import argparse
import sys
from typing import List, Optional

from celline.cli.commands import cmd_list, cmd_help, cmd_run, cmd_info, cmd_init, cmd_interactive, cmd_api, cmd_config, cmd_export


def create_parser() -> argparse.ArgumentParser:
    """Create the main argument parser."""
    parser = argparse.ArgumentParser(
        prog='celline',
        description='Celline - Single Cell Analysis Pipeline',
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    
    subparsers = parser.add_subparsers(dest='command', help='Available commands')
    
    # Init command
    init_parser = subparsers.add_parser('init', help='Initialize a new celline project')
    init_parser.add_argument('project_name', nargs='?', help='Project name')
    
    # List command
    list_parser = subparsers.add_parser('list', help='List all available functions')
    
    # Help command
    help_parser = subparsers.add_parser('help', help='Show help for a specific function')
    help_parser.add_argument('function_name', nargs='?', help='Function name to get help for')
    
    # Run command
    run_parser = subparsers.add_parser('run', help='Run a specific function')
    run_parser.add_argument('function_name', help='Function name to run')
    run_parser.add_argument('function_args', nargs='*', help='Arguments to pass to the function')
    run_parser.add_argument('--project-dir', '-p', default='.', help='Project directory (default: current directory)')
    run_parser.add_argument('--project-name', '-n', default='default', help='Project name (default: default)')
    
    # Info command
    info_parser = subparsers.add_parser('info', help='Show system information')
    
    # Config command
    config_parser = subparsers.add_parser('config', help='Configure celline settings')
    config_parser.add_argument('--system', choices=['multithreading', 'PBS'], help='Set execution system')
    config_parser.add_argument('--nthread', type=int, help='Set number of threads')
    config_parser.add_argument('--pbs-server', help='Set PBS server name (for PBS system)')
    
    # Interactive command
    interactive_parser = subparsers.add_parser('interactive', help='Launch interactive web interface')
    
    # API command (for testing)
    api_parser = subparsers.add_parser('api', help='Start API server only (for testing)')
    
    # Export command
    export_parser = subparsers.add_parser('export', help='Export data and reports')
    export_subparsers = export_parser.add_subparsers(dest='export_command', help='Export options')
    metareport_parser = export_subparsers.add_parser('metareport', help='Generate metadata report from samples.toml')
    metareport_parser.add_argument('--output', '-o', default='metadata_report.html', help='Output HTML file (default: metadata_report.html)')
    metareport_parser.add_argument('--project-dir', '-p', default='.', help='Project directory (default: current directory)')
    
    return parser


def main(argv: Optional[List[str]] = None) -> int:
    """Main CLI entry point."""
    parser = create_parser()
    
    if argv is None:
        argv = sys.argv[1:]
    
    # If no arguments provided, show help
    if not argv:
        parser.print_help()
        return 0
    
    args = parser.parse_args(argv)
    
    try:
        if args.command == 'init':
            cmd_init(args)
        elif args.command == 'list':
            cmd_list(args)
        elif args.command == 'help':
            cmd_help(args)
        elif args.command == 'run':
            # Check if the function is 'interactive'
            if hasattr(args, 'function_name') and args.function_name == 'interactive':
                cmd_interactive(args)
            else:
                cmd_run(args)
        elif args.command == 'info':
            cmd_info(args)
        elif args.command == 'config':
            cmd_config(args)
        elif args.command == 'interactive':
            cmd_interactive(args)
        elif args.command == 'api':
            cmd_api(args)
        elif args.command == 'export':
            cmd_export(args)
        else:
            parser.print_help()
            return 1
            
        return 0
        
    except KeyboardInterrupt:
        print("\nInterrupted by user")
        return 130
    except Exception as e:
        print(f"Error: {e}")
        return 1


if __name__ == '__main__':
    sys.exit(main())


"""
Enhanced CLI invoker that uses the new cli method from CellineFunction.
"""

import argparse
from typing import Type, Optional
from rich.console import Console

from celline.functions._base import CellineFunction
from celline.interfaces import Project

console = Console()


class EnhancedFunctionInvoker:
    """Enhanced invoker that uses the CLI interface of CellineFunction."""
    
    def __init__(self, func_class: Type[CellineFunction]):
        self.func_class = func_class
    
    def create_function_parser(self) -> argparse.ArgumentParser:
        """Create an argument parser for this specific function."""
        # Get function description using class methods without full instantiation
        try:
            # Try to create a minimal instance to get metadata
            temp_instance = self.func_class.__new__(self.func_class)
            # Initialize the ABC part to avoid abstract method errors
            temp_instance.__class__ = self.func_class
            
            description = temp_instance.get_description()
            command_name = temp_instance.register()
        except:
            # Fallback to defaults if instantiation fails
            description = self.func_class.__doc__ or f"{self.func_class.__name__} function"
            command_name = self.func_class.__name__.lower()
        
        parser = argparse.ArgumentParser(
            prog=f"celline run {command_name}",
            description=description,
            formatter_class=argparse.RawDescriptionHelpFormatter
        )
        
        # Let the function add its own arguments
        try:
            temp_instance.add_cli_args(parser)
        except:
            # If add_cli_args fails, continue without custom arguments
            pass
        
        # Add usage examples to help text
        try:
            examples = temp_instance.get_usage_examples()
            if examples:
                epilog = "Examples:\n" + "\n".join(f"  {example}" for example in examples)
                parser.epilog = epilog
        except:
            # If getting examples fails, continue without them
            pass
        
        return parser
    
    def get_help_text(self) -> str:
        """Get help text for this function."""
        parser = self.create_function_parser()
        return parser.format_help()
    
    def invoke(self, project: Project, cli_args: list[str] = None) -> Project:
        """
        Invoke the function using CLI interface.
        
        Args:
            project: The Project instance
            cli_args: CLI arguments to parse (if None, no arguments)
            
        Returns:
            Project: The project instance
        """
        try:
            # Create parser and parse arguments
            parser = self.create_function_parser()
            
            if cli_args:
                args = parser.parse_args(cli_args)
            else:
                args = argparse.Namespace()
            
            # Create function instance using CLI factory method
            # This allows functions to handle their own instantiation based on CLI args
            try:
                # Try to create a minimal instance first
                instance = self.func_class.__new__(self.func_class)
                instance.__class__ = self.func_class
                
                # Call the CLI method - it should handle its own instantiation
                console.print(f"[cyan]Executing {instance.register()}...[/cyan]")
                result = instance.cli(project, args)
                console.print(f"[green]✓ {instance.register()} completed successfully[/green]")
                
                return result
            except Exception as e:
                # If that fails, try to call the cli method statically
                temp_instance = self.func_class.__new__(self.func_class)
                temp_instance.__class__ = self.func_class
                result = temp_instance.cli(project, args)
                return result
            
        except SystemExit:
            # argparse calls sys.exit on error - catch and re-raise as regular exception
            raise ValueError("Invalid arguments provided")
        except Exception as e:
            console.print(f"[red]Error: {e}[/red]")
            raise


def invoke_function_enhanced(func_class: Type[CellineFunction], project: Project, cli_args: list[str] = None) -> Project:
    """Convenience function to invoke a CellineFunction using enhanced CLI interface."""
    invoker = EnhancedFunctionInvoker(func_class)
    return invoker.invoke(project, cli_args)


import argparse
import datetime
import logging
import os
from collections.abc import Callable
from pathlib import Path
from typing import TYPE_CHECKING, Final, Optional

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import polars as pl
import scanpy as sc
import seaborn as sns
import toml
from rich.console import Console

from celline.config import Config
from celline.DB.dev.model import SampleSchema
from celline.functions._base import CellineFunction
from celline.sample import SampleInfo, SampleResolver

if TYPE_CHECKING:
    from celline import Project

console = Console()


class Integrate(CellineFunction):
    """Integration function with support for multiple methods (scVI, Harmony)."""

    def __init__(
        self,
        filter_func: Callable[[SampleSchema], bool] | None = None,
        outfile_name: str | None = None,
        integration_method: str = "harmony",
        n_pcs: int = 50,
        batch_key: str = "sample",
        scvi_epochs: int = 200,
        scvi_early_stopping: bool = True,
        harmony_vars_use: list[str] | None = None,
        force_rerun: bool = False,
        verbose: bool = True,
        sample_list: list[str] | None = None,
    ) -> None:
        """Initialize the Integration class.

        Parameters
        ----------
        filter_func : Optional[Callable[[SampleSchema], bool]], default=None
            Function to filter samples to include in integration.
        outfile_name : Optional[str], default=None
            Output file name. If None, uses timestamp.
        integration_method : str, default="harmony"
            Integration method to use ("scvi" or "harmony").
        n_pcs : int, default=50
            Number of principal components to use.
        batch_key : str, default="sample"
            Column name in obs containing batch information.
        scvi_epochs : int, default=200
            Number of training epochs for scVI.
        scvi_early_stopping : bool, default=True
            Whether to use early stopping in scVI training.
        harmony_vars_use : Optional[list[str]], default=None
            Variables to use for Harmony correction. If None, uses batch_key.
        force_rerun : bool, default=False
            Whether to force rerun even if cached results exist.
        verbose : bool, default=True
            Whether to show detailed progress.
        sample_list : Optional[list[str]], default=None
            List of specific sample IDs to integrate. If None, uses all available samples.

        """
        self.filter_func = filter_func
        self.integration_method = integration_method.lower()
        self.n_pcs = n_pcs
        self.batch_key = batch_key
        self.scvi_epochs = scvi_epochs
        self.scvi_early_stopping = scvi_early_stopping
        self.harmony_vars_use = harmony_vars_use if harmony_vars_use else [batch_key]
        self.force_rerun = force_rerun
        self.verbose = verbose
        self.sample_list = sample_list

        # Setup output path
        if outfile_name is None:
            timestamp = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
            self.outfile_name = f"integrated_{self.integration_method}_{timestamp}"
        else:
            self.outfile_name = outfile_name

        # Create method-specific output directory
        self.output_dir = Path(Config.PROJ_ROOT) / "integration" / self.integration_method
        # Note: output_path is now managed in _save_results with structured directories

        # Validate integration method
        if self.integration_method not in ["scvi", "harmony"]:
            raise ValueError(f"Integration method must be 'scvi' or 'harmony', got '{self.integration_method}'")

    def register(self) -> str:
        return "integrate"

    def _get_cache_path(self) -> Path:
        """Get path to the integration cache TOML file."""
        return self.output_dir / "integrate.toml"

    def _generate_sample_key(self, sample_list: list[str]) -> str:
        """Generate a unique key for a sample combination."""
        return "_".join(sorted(sample_list))

    def _load_cache(self) -> dict:
        """Load existing integration cache."""
        cache_path = self._get_cache_path()
        if cache_path.exists():
            try:
                return toml.load(cache_path)
            except Exception as e:
                console.print(f"[yellow]⚠ Warning: Could not load cache file: {e}[/yellow]")
                return {}
        return {}

    def _save_cache(self, cache_data: dict):
        """Save integration cache to TOML file."""
        cache_path = self._get_cache_path()
        try:
            with open(cache_path, "w") as f:
                toml.dump(cache_data, f)
        except Exception as e:
            console.print(f"[yellow]⚠ Warning: Could not save cache file: {e}[/yellow]")

    def _check_cached_integration(self, sample_ids: list[str]) -> str | None:
        """Check if this sample combination has been integrated before."""
        cache = self._load_cache()
        sample_key = self._generate_sample_key(sample_ids)

        if sample_key in cache:
            cached_entry = cache[sample_key]
            cached_dir = Path(cached_entry.get("output_dir", ""))

            # Verify that the cached directory and model still exist
            if cached_dir.exists():
                if self.integration_method == "scvi":
                    model_path = cached_dir / "models" / "scvi_model"
                    if model_path.exists():
                        console.print(f"[green]✓ Found cached integration for samples: {', '.join(sample_ids)}[/green]")
                        console.print(f"[green]Using cached results from: {cached_dir}[/green]")
                        return str(cached_dir)
                else:
                    # For harmony, just check if h5ad file exists
                    h5ad_path = cached_dir / "data" / f"{cached_dir.name}.h5ad"
                    if h5ad_path.exists():
                        console.print(f"[green]✓ Found cached integration for samples: {', '.join(sample_ids)}[/green]")
                        console.print(f"[green]Using cached results from: {cached_dir}[/green]")
                        return str(cached_dir)

        return None

    def _add_to_cache(self, sample_ids: list[str], output_dir: str):
        """Add a new integration to the cache."""
        cache = self._load_cache()
        sample_key = self._generate_sample_key(sample_ids)

        cache[sample_key] = {
            "samples": sample_ids,
            "output_dir": output_dir,
            "integration_method": self.integration_method,
            "timestamp": datetime.datetime.now().isoformat(),
            "n_samples": len(sample_ids),
        }

        self._save_cache(cache)
        console.print(f"[dim]Added integration to cache: {sample_key}[/dim]")

    def _apply_celltype_mapping_to_integrated_data(self, adata: sc.AnnData, target_samples: list[SampleInfo], logger: logging.Logger):
        """Apply celltype mapping to already integrated data."""
        try:
            # Initialize cell_type column with 'Unknown'
            adata.obs["cell_type"] = "Unknown"

            # Process each sample
            for sample_info in target_samples:
                sample_key = sample_info.schema.key
                data_sample_dir = sample_info.path.data_sample
                celltype_path = f"{data_sample_dir}/celltype_predicted.tsv"

                if not os.path.exists(celltype_path):
                    logger.warning(f"celltype_predicted.tsv not found for {sample_key}")
                    continue

                try:
                    # Load celltype predictions
                    celltype_df = pd.read_csv(celltype_path, sep="\t")
                    if "cell" not in celltype_df.columns:
                        logger.warning(f"No 'cell' column in celltype_predicted.tsv for {sample_key}")
                        continue

                    # Get celltype column (scpred_prediction or cell_type_cluster_weighted)
                    celltype_col = "cell_type_cluster_weighted" if "cell_type_cluster_weighted" in celltype_df.columns else "scpred_prediction"

                    # Get cells for this sample from integrated data
                    sample_mask = adata.obs["sample"] == sample_key
                    sample_cells = adata.obs.index[sample_mask]

                    if len(sample_cells) == 0:
                        logger.warning(f"No cells found for sample {sample_key} in integrated data")
                        continue

                    # Try direct mapping first
                    celltype_dict = dict(zip(celltype_df["cell"], celltype_df[celltype_col], strict=False))
                    mapped_types = pd.Series(sample_cells).map(celltype_dict)
                    direct_mapped_count = mapped_types.notna().sum()

                    if direct_mapped_count > len(sample_cells) * 0.5:
                        # Direct mapping worked well
                        adata.obs.loc[sample_cells, "cell_type"] = mapped_types.fillna("Unknown").values
                        logger.info(f"Direct mapping successful for {sample_key}: {direct_mapped_count}/{len(sample_cells)} cells")
                    else:
                        # Use position-based mapping
                        logger.info(f"Using position-based mapping for {sample_key} ({direct_mapped_count}/{len(sample_cells)} direct matches)")

                        # Sort both datasets
                        celltype_df = celltype_df.sort_values("cell")
                        sample_cells_sorted = sorted(sample_cells)

                        # Map by position (assume celltype data is a subset in order)
                        for i, celltype_value in enumerate(celltype_df[celltype_col]):
                            if i < len(sample_cells_sorted):
                                cell_name = sample_cells_sorted[i]
                                adata.obs.loc[cell_name, "cell_type"] = celltype_value

                        position_mapped_count = (adata.obs.loc[sample_cells, "cell_type"] != "Unknown").sum()
                        logger.info(f"Position-based mapping for {sample_key}: {position_mapped_count}/{len(sample_cells)} cells")

                except Exception as e:
                    logger.warning(f"Failed to apply celltype mapping for {sample_key}: {e}")

            # Report overall mapping success
            total_mapped = (adata.obs["cell_type"] != "Unknown").sum()
            total_cells = len(adata.obs)
            logger.info(f"Overall celltype mapping: {total_mapped}/{total_cells} cells ({total_mapped / total_cells * 100:.1f}%)")

        except Exception as e:
            logger.error(f"Failed to apply celltype mapping to integrated data: {e}")
            # Ensure cell_type column exists even if mapping fails
            if "cell_type" not in adata.obs.columns:
                adata.obs["cell_type"] = "Unknown"

    def _load_cached_and_plot(self, cached_dir: str, target_samples: list[SampleInfo], logger: logging.Logger, project: "Project"):
        """Load cached integration results and regenerate plots only."""
        try:
            cached_path = Path(cached_dir)
            data_dir = cached_path / "data"
            h5ad_file = data_dir / f"{cached_path.name}.h5ad"

            if not h5ad_file.exists():
                console.print(f"[red]❌ Cached h5ad file not found at {h5ad_file}[/red]")
                console.print("[yellow]Falling back to full integration...[/yellow]")
                return None

            # Load cached AnnData
            console.print(f"[dim]Loading cached data from {h5ad_file}...[/dim]")
            adata_integrated = sc.read_h5ad(h5ad_file)

            # Apply celltype mapping to cached data if not already present
            if "cell_type" not in adata_integrated.obs.columns:
                console.print("[dim]Adding celltype information to cached data...[/dim]")
                self._apply_celltype_mapping_to_integrated_data(adata_integrated, target_samples, logger)

            # Regenerate figures
            figures_dir = cached_path / "figures"
            figures_dir.mkdir(exist_ok=True)

            console.print("[dim]Regenerating integration plots...[/dim]")
            self._generate_integration_plots(adata_integrated, figures_dir, logger)

            console.print("[green]✅ Cached integration loaded successfully![/green]")
            console.print(f"[green]Plots regenerated at: {figures_dir}[/green]")

            return project

        except Exception as e:
            console.print(f"[red]❌ Failed to load cached integration: {e}[/red]")
            console.print("[yellow]Falling back to full integration...[/yellow]")
            logger.warning(f"Failed to load cached integration: {e}")
            return None

    def call(self, project: "Project"):
        """Main integration pipeline.

        Parameters
        ----------
        project : Project
            Celline project object.

        Returns
        -------
        Project
            Input project object.

        """
        console.print(f"[cyan]Starting integration with {self.integration_method.upper()} method...[/cyan]")

        # Create output directories
        self.output_dir.mkdir(parents=True, exist_ok=True)
        (self.output_dir / "logs").mkdir(exist_ok=True)

        # Setup logging
        log_file = self.output_dir / "logs" / f"{self.outfile_name}.log"
        logging.basicConfig(
            level=logging.INFO if self.verbose else logging.WARNING,
            format="%(asctime)s - %(levelname)s - %(message)s",
            handlers=[logging.FileHandler(log_file), logging.StreamHandler() if self.verbose else logging.NullHandler()],
        )
        logger = logging.getLogger(__name__)

        try:
            # Check for cached integration first (based on requested samples, not discovered samples)
            cached_dir = None
            if not self.force_rerun and self.sample_list:
                console.print(f"[dim]Checking cache for requested samples: {self.sample_list}[/dim]")
                cached_dir = self._check_cached_integration(self.sample_list)
                console.print(f"[dim]Cache check result: {cached_dir}[/dim]")

            if cached_dir:
                # For cached integration, create minimal target_samples for the cache loading
                console.print(f"[cyan]Found cached integration for requested samples: {', '.join(self.sample_list)}[/cyan]")
                # Create dummy target_samples for celltype mapping
                target_samples = []
                all_samples = SampleResolver.samples()
                console.print(f"[dim]SampleResolver has {len(all_samples)} samples: {list(all_samples.keys())}[/dim]")

                for sample_id in self.sample_list:
                    # We need to create minimal SampleInfo objects or work around this
                    # For now, try to get sample info from SampleResolver if available
                    if sample_id in all_samples:
                        target_samples.append(all_samples[sample_id])
                        console.print(f"[dim]Found sample info for {sample_id}[/dim]")
                    else:
                        console.print(f"[dim]No sample info found for {sample_id} in SampleResolver[/dim]")

                console.print("[cyan]Loading cached integration and regenerating plots...[/cyan]")
                result = self._load_cached_and_plot(cached_dir, target_samples, logger, project)
                if result is not None:
                    return result
                console.print("[cyan]Cached loading failed, continuing with sample discovery...[/cyan]")

            # Collect target samples
            console.print(f"[dim]Attempting to collect target samples with sample_list: {self.sample_list}[/dim]")
            target_samples = self._collect_target_samples()
            console.print(f"[dim]Collected {len(target_samples)} target samples[/dim]")
            if not target_samples:
                console.print("[red]❌ No valid samples found for integration[/red]")
                return project

            sample_ids = [info.schema.key for info in target_samples]
            console.print(f"[green]Found {len(target_samples)} samples for integration: {', '.join(sample_ids)}[/green]")

            # Check for celltype predictions
            self._check_celltype_predictions(target_samples, logger)

            # Check for cached integration again (unless force_rerun is True) - this time based on discovered samples
            if not cached_dir and not self.force_rerun:
                cached_dir = self._check_cached_integration(sample_ids)

            if cached_dir:
                # Load cached integration and regenerate plots only
                console.print("[cyan]Loading cached integration and regenerating plots...[/cyan]")
                result = self._load_cached_and_plot(cached_dir, target_samples, logger, project)
                if result is not None:
                    return result
                # If loading cached results failed, continue with full integration
                console.print("[cyan]Falling back to full integration...[/cyan]")
            else:
                console.print("[cyan]No cached integration found, performing full integration...[/cyan]")

            # Load and combine data
            adata_combined = self._load_and_combine_samples(target_samples, logger)

            # Perform integration
            if self.integration_method == "scvi":
                adata_integrated = self._integrate_scvi(adata_combined, logger)
            elif self.integration_method == "harmony":
                adata_integrated = self._integrate_harmony(adata_combined, logger)

            # Save results
            self._save_results(adata_integrated, logger)

            # Add to cache for future reuse
            output_path = str(self.output_dir / self.outfile_name)
            self._add_to_cache(sample_ids, output_path)

            console.print("[green]✅ Integration completed successfully![/green]")
            console.print(f"[green]Results saved to: {self.output_dir / self.outfile_name}[/green]")

        except Exception as e:
            console.print(f"[red]❌ Integration failed: {e}[/red]")
            logger.error(f"Integration failed: {e}", exc_info=True)
            raise

        return project

    def _collect_target_samples(self) -> list[SampleInfo]:
        """Collect target samples based on filter function."""
        target_samples: list[SampleInfo] = []

        all_samples = SampleResolver.samples()
        console.print(f"[dim]SampleResolver found {len(all_samples)} samples: {list(all_samples.keys())}[/dim]")

        for info in all_samples.values():
            console.print(f"[dim]Checking sample {info.schema.key}: is_counted={info.path.is_counted}[/dim]")
            if info.path.is_counted:
                # Apply sample list filter if specified
                if self.sample_list is not None and info.schema.key not in self.sample_list:
                    console.print(f"[dim]Sample {info.schema.key} not in requested list, skipping[/dim]")
                    continue

                if self.filter_func is None:
                    add = True
                else:
                    add = self.filter_func(info.schema)
                if add:
                    console.print(f"[dim]Adding sample {info.schema.key} to target list[/dim]")
                    target_samples.append(info)
                else:
                    console.print(f"[dim]Sample {info.schema.key} filtered out by filter_func[/dim]")
            else:
                console.print(f"[yellow]⚠ Warning: Sample {info.schema.key} is not counted or preprocessed yet[/yellow]")

        return target_samples

    def _check_celltype_predictions(self, target_samples: list[SampleInfo], logger: logging.Logger):
        """Check if celltype_predicted.tsv exists for each sample."""
        missing_celltype_samples = []

        for sample_info in target_samples:
            celltype_path = f"{sample_info.path.data_sample}/celltype_predicted.tsv"
            if not os.path.exists(celltype_path):
                missing_celltype_samples.append(sample_info.schema.key)

        if missing_celltype_samples:
            console.print(f"[yellow]⚠ Warning: celltype_predicted.tsv not found for samples: {', '.join(missing_celltype_samples)}[/yellow]")
            console.print("[yellow]Cell type integration will be limited for these samples[/yellow]")
            logger.warning(f"Missing celltype predictions for: {missing_celltype_samples}")
        else:
            console.print("[green]✓ All samples have celltype predictions available[/green]")
            logger.info("All samples have celltype predictions available")

    def _load_and_combine_samples(self, target_samples: list[SampleInfo], logger: logging.Logger) -> sc.AnnData:
        """Load individual samples and combine them."""
        logger.info(f"Loading {len(target_samples)} samples...")
        console.print("[dim]Loading and combining samples...[/dim]")

        adata_list = []

        for i, sample_info in enumerate(target_samples, 1):
            console.print(f"[dim]Loading sample {i}/{len(target_samples)}: {sample_info.schema.key}[/dim]")

            # Load count matrix
            count_matrix_path = f"{sample_info.path.resources_sample_counted}/outs/filtered_feature_bc_matrix.h5"

            if not os.path.exists(count_matrix_path):
                logger.warning(f"Count matrix not found for {sample_info.schema.key}: {count_matrix_path}")
                continue

            adata = sc.read_10x_h5(count_matrix_path)
            adata.var_names_make_unique()

            # Add sample metadata and fix cell indexing to start from 1
            adata.obs["sample"] = sample_info.schema.key
            adata.obs["project"] = sample_info.schema.parent or "Unknown"

            # Use polars to create proper cell indexing starting from 1
            obs_df = (
                pl.DataFrame(adata.obs)
                .with_columns(
                    pl.format(
                        "{}_{}",
                        pl.col("sample"),
                        (
                            pl.col("sample")  # Count by sample
                            .cum_count()
                            .over("sample")
                        ),  # 1-based indexing
                    ).alias("cell"),
                )
                .to_pandas()
                .set_index("cell")
            )

            adata.obs = obs_df
            adata.obs_names = adata.obs.index

            # Load additional metadata if available
            self._load_additional_metadata(adata, sample_info, logger)

            adata_list.append(adata)
            logger.info(f"Loaded {sample_info.schema.key}: {adata.n_obs} cells × {adata.n_vars} genes")

        if not adata_list:
            raise ValueError("No valid samples could be loaded")

        # Combine all samples
        console.print("[dim]Concatenating samples...[/dim]")
        adata_combined = sc.concat(adata_list, join="outer")

        # Make variable names unique
        adata_combined.var_names_make_unique()

        logger.info(f"Combined data: {adata_combined.n_obs} cells × {adata_combined.n_vars} genes")
        console.print(f"[green]Combined data: {adata_combined.n_obs} cells × {adata_combined.n_vars} genes[/green]")

        return adata_combined

    def _load_additional_metadata(self, adata: sc.AnnData, sample_info: SampleInfo, logger: logging.Logger):
        """Load additional metadata files (cell type predictions, QC metrics, etc.)."""
        data_sample_dir = sample_info.path.data_sample

        # Load cell type predictions
        celltype_path = f"{data_sample_dir}/celltype_predicted.tsv"
        if os.path.exists(celltype_path):
            try:
                celltype_df = pd.read_csv(celltype_path, sep="\t")
                if "cell" in celltype_df.columns:
                    # Use cell_type_cluster_weighted if available, otherwise fall back to scpred_prediction
                    celltype_col = "cell_type_cluster_weighted" if "cell_type_cluster_weighted" in celltype_df.columns else "scpred_prediction"

                    # Create mapping dictionary
                    celltype_dict = dict(zip(celltype_df["cell"], celltype_df.get(celltype_col, "Unknown"), strict=False))

                    # Try direct mapping first
                    adata.obs["cell_type"] = adata.obs.index.map(celltype_dict)

                    # If direct mapping failed (many NaNs), try alternative mapping strategies
                    mapped_count = adata.obs["cell_type"].notna().sum()
                    total_count = len(adata.obs)

                    if mapped_count < total_count * 0.5:  # Less than 50% mapped
                        logger.warning(f"Direct cell mapping resulted in only {mapped_count}/{total_count} cells mapped")

                        # Alternative strategy: map by position if cell numbers are sequential
                        # Extract cell numbers from both datasets
                        sample_key = sample_info.schema.key

                        # Get cell numbers from celltype data (e.g., GSM3934448_3 -> 3)
                        celltype_df["cell_num"] = celltype_df["cell"].str.extract(f"{sample_key}_(\\d+)").astype(int)
                        celltype_df = celltype_df.sort_values("cell_num")

                        # Get cell numbers from adata (e.g., GSM3934448_1 -> 1)
                        adata_cell_nums = pd.Series(adata.obs.index).str.extract(f"{sample_key}_(\\d+)")[0].astype(int)

                        # Create position-based mapping if the celltype data appears to be a subset
                        if len(celltype_df) <= len(adata.obs):
                            # Map by relative position, accounting for potential offset
                            min_celltype_num = celltype_df["cell_num"].min()
                            max_celltype_num = celltype_df["cell_num"].max()

                            # Create mapping based on the assumption that celltype data
                            # corresponds to cells in order, possibly with an offset
                            for i, (_, row) in enumerate(celltype_df.iterrows()):
                                if i < len(adata.obs):
                                    cell_name = adata.obs.index[i]
                                    adata.obs.loc[cell_name, "cell_type"] = row[celltype_col]

                        logger.info("Applied position-based mapping for celltype predictions")

                    # Fill remaining NaNs with 'Unknown'
                    adata.obs["cell_type"] = adata.obs["cell_type"].fillna("Unknown")

                    final_mapped = adata.obs["cell_type"].ne("Unknown").sum()
                    logger.info(f"Successfully mapped cell types for {final_mapped}/{total_count} cells using column: {celltype_col}")

            except Exception as e:
                logger.warning(f"Could not load cell type predictions for {sample_info.schema.key}: {e}")

        # Load QC metrics
        qc_path = f"{data_sample_dir}/qc_matrix.tsv"
        if os.path.exists(qc_path):
            try:
                qc_df = pd.read_csv(qc_path, sep="\t")
                # Map QC metrics by barcode or cell name
                for col in qc_df.columns:
                    if col not in ["cell", "barcodes"]:
                        if "cell" in qc_df.columns:
                            qc_dict = dict(zip(qc_df["cell"], qc_df[col], strict=False))
                            adata.obs[col] = adata.obs.index.map(qc_dict)
                logger.debug(f"Loaded QC metrics for {sample_info.schema.key}")
            except Exception as e:
                logger.warning(f"Could not load QC metrics for {sample_info.schema.key}: {e}")

    def _integrate_scvi(self, adata: sc.AnnData, logger: logging.Logger) -> sc.AnnData:
        """Integrate data using scVI method."""
        try:
            import scvi
            import torch
        except ImportError:
            raise ImportError("scvi-tools package is required for scVI integration. Install with: pip install scvi-tools")

        logger.info("Starting scVI integration...")
        console.print("[dim]Performing scVI integration...[/dim]")

        # Setup scVI model
        logger.info("Setting up scVI model...")
        scvi.model.SCVI.setup_anndata(adata, batch_key=self.batch_key)

        # Create and train model
        model = scvi.model.SCVI(adata)
        logger.info(f"Training scVI model for {self.scvi_epochs} epochs...")
        console.print(f"[dim]Training scVI model ({self.scvi_epochs} epochs)...[/dim]")

        model.train(max_epochs=self.scvi_epochs, early_stopping=self.scvi_early_stopping, accelerator="cpu")

        # Get latent representation
        logger.info("Extracting latent representation...")
        adata.obsm["X_scvi"] = model.get_latent_representation()

        # Compute neighbors and UMAP (scVI latent representation doesn't need n_pcs)
        sc.pp.neighbors(adata, use_rep="X_scvi")
        sc.tl.umap(adata)
        sc.tl.leiden(adata, resolution=1.0, key_added="leiden_scvi")

        # Save scVI model
        model_save_dir = self.output_dir / self.outfile_name / "models"
        model_save_dir.mkdir(parents=True, exist_ok=True)
        model_save_path = model_save_dir / "scvi_model"
        logger.info(f"Saving scVI model to {model_save_path}")
        console.print(f"[dim]Saving scVI model to {model_save_path}...[/dim]")
        model.save(model_save_path, overwrite=True)

        logger.info("scVI integration completed")
        console.print("[green]✓ scVI integration completed[/green]")

        return adata

    def _integrate_harmony(self, adata: sc.AnnData, logger: logging.Logger) -> sc.AnnData:
        """Integrate data using Harmony method."""
        try:
            import harmonypy as hm
        except ImportError:
            raise ImportError("harmonypy package is required for Harmony integration. Install with: pip install harmonypy")

        logger.info("Starting Harmony integration...")
        console.print("[dim]Performing Harmony integration...[/dim]")

        # Basic preprocessing
        logger.info("Performing basic preprocessing...")
        sc.pp.normalize_total(adata, target_sum=1e4)
        sc.pp.log1p(adata)
        sc.pp.highly_variable_genes(adata, n_top_genes=2000, batch_key=self.batch_key)
        adata.raw = adata
        adata = adata[:, adata.var.highly_variable]
        sc.pp.scale(adata, max_value=10)

        # PCA
        logger.info(f"Computing PCA with {self.n_pcs} components...")
        sc.tl.pca(adata, svd_solver="arpack", n_comps=self.n_pcs)

        # Prepare data for Harmony
        data_mat = adata.obsm["X_pca"].T  # Harmony expects genes x cells
        meta_data = adata.obs[self.harmony_vars_use].copy()

        # Run Harmony
        logger.info(f"Running Harmony with vars_use: {self.harmony_vars_use}")
        console.print("[dim]Running Harmony correction...[/dim]")

        ho = hm.run_harmony(data_mat, meta_data, self.harmony_vars_use)

        # Store corrected embedding
        adata.obsm["X_harmony"] = ho.Z_corr.T

        # Compute neighbors and UMAP using corrected embedding
        sc.pp.neighbors(adata, use_rep="X_harmony", n_pcs=self.n_pcs)
        sc.tl.umap(adata)
        sc.tl.leiden(adata, resolution=1.0, key_added="leiden_harmony")

        logger.info("Harmony integration completed")
        console.print("[green]✓ Harmony integration completed[/green]")

        return adata

    def _save_results(self, adata: sc.AnnData, logger: logging.Logger):
        """Save integrated results and generate plots."""
        main_output_dir = self.output_dir / self.outfile_name
        logger.info(f"Saving integrated data to {main_output_dir}")
        console.print(f"[dim]Saving results to {main_output_dir}...[/dim]")

        # Create structured output directories
        data_dir = self.output_dir / self.outfile_name / "data"
        figures_dir = self.output_dir / self.outfile_name / "figures"
        data_dir.mkdir(parents=True, exist_ok=True)
        figures_dir.mkdir(parents=True, exist_ok=True)

        # Save the main integrated data with only specified columns
        main_output_path = data_dir / f"{self.outfile_name}.h5ad"

        # Create a copy with only the required obs columns
        adata_save = adata.copy()

        # Define required columns based on integration method
        if self.integration_method == "scvi":
            required_columns = ["sample", "project", "_scvi_batch", "_scvi_labels", "leiden_scvi"]
        else:  # harmony
            required_columns = ["sample", "project", "leiden"]

        # Keep only required columns that exist in the data
        existing_columns = [col for col in required_columns if col in adata_save.obs.columns]
        if not existing_columns:
            logger.warning("None of the required columns found in data, saving with current obs")
            existing_columns = list(adata_save.obs.columns)

        # Filter obs to only include required columns
        adata_save.obs = adata_save.obs[existing_columns].copy()

        # Ensure cell names are in the index
        if "cell" in adata_save.obs.columns:
            adata_save.obs_names = adata_save.obs["cell"]
            adata_save.obs = adata_save.obs.drop("cell", axis=1)

        adata_save.write_h5ad(main_output_path)
        logger.info(f"Integrated data saved to {main_output_path} with columns: {existing_columns}")

        # Generate and save plots
        console.print("[dim]Generating integration plots...[/dim]")
        self._generate_integration_plots(adata, figures_dir, logger)

        # Save summary statistics
        summary_path = data_dir / f"{self.outfile_name}_summary.txt"
        with open(summary_path, "w") as f:
            f.write("Integration Summary\n")
            f.write("==================\n")
            f.write(f"Method: {self.integration_method.upper()}\n")
            f.write(f"Total cells: {adata.n_obs}\n")
            f.write(f"Total genes: {adata.n_vars}\n")
            f.write(f"Samples: {len(adata.obs[self.batch_key].unique())}\n")
            f.write(f"Projects: {len(adata.obs['project'].unique())}\n")
            f.write(f"Batch key: {self.batch_key}\n")
            f.write(f"Number of PCs: {self.n_pcs}\n")
            if self.integration_method == "scvi":
                f.write(f"scVI epochs: {self.scvi_epochs}\n")
                f.write(f"Early stopping: {self.scvi_early_stopping}\n")
            elif self.integration_method == "harmony":
                f.write(f"Harmony vars: {', '.join(self.harmony_vars_use)}\n")

        logger.info(f"Summary saved to {summary_path}")
        console.print(f"[green]✓ Results saved to: {self.output_dir / self.outfile_name}[/green]")

    def _generate_integration_plots(self, adata: sc.AnnData, figures_dir: Path, logger: logging.Logger):
        """Generate comprehensive integration plots."""
        logger.info("Generating integration plots...")

        # Set up matplotlib parameters
        plt.style.use("default")
        sns.set_palette("husl")
        sc.settings.figdir = str(figures_dir)

        # Set matplotlib parameters directly to avoid recursion
        import matplotlib as mpl

        mpl.rcParams["figure.dpi"] = 300
        mpl.rcParams["figure.facecolor"] = "white"
        mpl.rcParams["savefig.format"] = "png"

        try:
            # Determine integration key for plots
            integration_key = "X_scvi" if self.integration_method == "scvi" else "X_harmony"
            leiden_key = f"leiden_{self.integration_method}"

            # 1. UMAP colored by clusters
            if leiden_key in adata.obs.columns:
                logger.debug("Generating cluster UMAP...")
                sc.pl.umap(adata, color=leiden_key, legend_loc="on data", legend_fontsize=8, frameon=False, save=f"_clusters_{self.integration_method}.png", show=False)

            # 2. UMAP colored by sample (batch effect visualization)
            logger.debug("Generating sample UMAP...")
            sc.pl.umap(adata, color=self.batch_key, legend_loc="right margin", frameon=False, save=f"_samples_{self.integration_method}.png", show=False)

            # 3. UMAP colored by project
            if "project" in adata.obs.columns and len(adata.obs["project"].unique()) > 1:
                logger.debug("Generating project UMAP...")
                sc.pl.umap(adata, color="project", legend_loc="right margin", frameon=False, save=f"_projects_{self.integration_method}.png", show=False)

            # 4. UMAP colored by cell type (if available)
            if "cell_type" in adata.obs.columns:
                logger.debug("Generating cell type UMAP...")
                sc.pl.umap(adata, color="cell_type", legend_loc="right margin", frameon=False, save=f"_celltypes_{self.integration_method}.png", show=False)

            # 5. QC metrics on UMAP
            qc_metrics = ["total_counts", "n_genes_by_counts", "pct_counts_mt"]
            available_qc = [metric for metric in qc_metrics if metric in adata.obs.columns]

            if available_qc:
                logger.debug("Generating QC metrics UMAP...")
                sc.pl.umap(adata, color=available_qc, ncols=2, frameon=False, save=f"_qc_metrics_{self.integration_method}.png", show=False)

            # 6. Integration quality plots
            self._generate_integration_quality_plots(adata, figures_dir, logger)

            # 7. Cluster composition plots
            if leiden_key in adata.obs.columns:
                self._generate_cluster_composition_plots(adata, figures_dir, leiden_key, logger)

            # 8. Comprehensive summary plots
            self._generate_comprehensive_summary_plots(adata, figures_dir, logger)

            # 9. Cell type analysis plots (if cell types are available)
            if "cell_type" in adata.obs.columns and adata.obs["cell_type"].ne("Unknown").any():
                self._generate_celltype_analysis_plots(adata, figures_dir, logger)

            # 10. scib metrics evaluation (disabled for performance)
            # if (len(adata.obs[self.batch_key].unique()) > 1 and
            #     'cell_type' in adata.obs.columns and
            #     adata.obs['cell_type'].ne('Unknown').any()):
            #     self._generate_scib_metrics(adata, figures_dir, logger)

            logger.info(f"Integration plots saved to {figures_dir}")
            console.print("[green]✓ Integration plots generated[/green]")

        except Exception as e:
            logger.warning(f"Could not generate some plots: {e}")
            console.print(f"[yellow]⚠ Warning: Some plots could not be generated: {e}[/yellow]")

    def _generate_integration_quality_plots(self, adata: sc.AnnData, figures_dir: Path, logger: logging.Logger):
        """Generate plots to assess integration quality."""
        try:
            logger.info("Generating integration quality assessment plots...")

            # Create integration quality visualization
            fig, axes = plt.subplots(2, 2, figsize=(16, 12))
            fig.suptitle(f"Integration Quality Assessment - {self.integration_method.upper()}", fontsize=16)

            integration_key = "X_scvi" if self.integration_method == "scvi" else "X_harmony"

            # Plot 1: UMAP colored by sample (batch effect assessment)
            ax1 = axes[0, 0]
            sample_colors = adata.obs[self.batch_key].astype("category").cat.codes
            scatter1 = ax1.scatter(adata.obsm["X_umap"][:, 0], adata.obsm["X_umap"][:, 1], c=sample_colors, s=0.5, alpha=0.7, cmap="tab10")
            ax1.set_title("Integration Quality: Sample Mixing")
            ax1.set_xlabel("UMAP 1")
            ax1.set_ylabel("UMAP 2")
            ax1.grid(True, alpha=0.3)

            # Add sample legend
            unique_samples = adata.obs[self.batch_key].unique()
            for i, sample in enumerate(unique_samples):
                ax1.scatter([], [], c=plt.cm.tab10(i), label=sample, s=20)
            ax1.legend(bbox_to_anchor=(1.05, 1), loc="upper left", fontsize=8)

            # Plot 2: UMAP colored by clusters
            ax2 = axes[0, 1]
            leiden_key = f"leiden_{self.integration_method}"
            if leiden_key in adata.obs.columns:
                cluster_colors = adata.obs[leiden_key].astype("category").cat.codes
                scatter2 = ax2.scatter(adata.obsm["X_umap"][:, 0], adata.obsm["X_umap"][:, 1], c=cluster_colors, s=0.5, alpha=0.7, cmap="tab20")
                ax2.set_title("Cluster Structure")
            else:
                ax2.text(0.5, 0.5, "No cluster information", ha="center", va="center", transform=ax2.transAxes)
                ax2.set_title("Cluster Structure (N/A)")
            ax2.set_xlabel("UMAP 1")
            ax2.set_ylabel("UMAP 2")
            ax2.grid(True, alpha=0.3)

            # Plot 3: Embedding space comparison (integrated vs raw if available)
            ax3 = axes[1, 0]
            if integration_key in adata.obsm:
                # Show first 2 dimensions of integrated embedding
                integrated_embed = adata.obsm[integration_key][:, :2]
                scatter3 = ax3.scatter(integrated_embed[:, 0], integrated_embed[:, 1], c=sample_colors, s=0.5, alpha=0.7, cmap="tab10")
                ax3.set_title(f"{self.integration_method.upper()} Embedding Space (2D)")
                ax3.set_xlabel(f"{integration_key} 1")
                ax3.set_ylabel(f"{integration_key} 2")
            else:
                ax3.text(0.5, 0.5, "Integration embedding not available", ha="center", va="center", transform=ax3.transAxes)
                ax3.set_title("Integration Embedding (N/A)")
            ax3.grid(True, alpha=0.3)

            # Plot 4: Sample distribution statistics
            ax4 = axes[1, 1]
            sample_counts = adata.obs[self.batch_key].value_counts()
            bars = ax4.bar(range(len(sample_counts)), sample_counts.values, color=[plt.cm.tab10(i) for i in range(len(sample_counts))])
            ax4.set_title("Cell Count per Sample")
            ax4.set_xlabel("Sample")
            ax4.set_ylabel("Number of Cells")
            ax4.set_xticks(range(len(sample_counts)))
            ax4.set_xticklabels(sample_counts.index, rotation=45, ha="right")
            ax4.grid(True, alpha=0.3, axis="y")

            # Add value labels on bars
            for i, bar in enumerate(bars):
                ax4.text(bar.get_x() + bar.get_width() / 2, bar.get_height() + max(sample_counts) * 0.01, f"{sample_counts.values[i]}", ha="center", va="bottom", fontsize=10)

            plt.tight_layout()
            plt.savefig(figures_dir / f"integration_quality_{self.integration_method}.png", dpi=300, bbox_inches="tight")
            plt.close()

            logger.info("Integration quality plots generated successfully")

        except Exception as e:
            logger.warning(f"Could not generate integration quality plots: {e}")
            import traceback

            logger.debug(f"Traceback: {traceback.format_exc()}")

    def _generate_cluster_composition_plots(self, adata: sc.AnnData, figures_dir: Path, leiden_key: str, logger: logging.Logger):
        """Generate cluster composition analysis plots."""
        try:
            # Cluster composition by sample
            cluster_sample_counts = adata.obs.groupby([leiden_key, self.batch_key]).size().unstack(fill_value=0)

            # Proportional stacked bar plot
            fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(16, 6))
            fig.suptitle(f"Cluster Composition Analysis - {self.integration_method.upper()}", fontsize=14)

            # Absolute counts
            cluster_sample_counts.plot(kind="bar", stacked=True, ax=ax1, colormap="tab20", width=0.8)
            ax1.set_title("Cluster Composition (Absolute Counts)")
            ax1.set_xlabel("Cluster")
            ax1.set_ylabel("Number of Cells")
            ax1.legend(title="Sample", bbox_to_anchor=(1.05, 1), loc="upper left")
            ax1.tick_params(axis="x", rotation=45)

            # Proportional
            cluster_sample_props = cluster_sample_counts.div(cluster_sample_counts.sum(axis=1), axis=0)
            cluster_sample_props.plot(kind="bar", stacked=True, ax=ax2, colormap="tab20", width=0.8)
            ax2.set_title("Cluster Composition (Proportions)")
            ax2.set_xlabel("Cluster")
            ax2.set_ylabel("Proportion of Cells")
            ax2.legend(title="Sample", bbox_to_anchor=(1.05, 1), loc="upper left")
            ax2.tick_params(axis="x", rotation=45)

            plt.tight_layout()
            plt.savefig(figures_dir / f"cluster_composition_{self.integration_method}.png", dpi=300, bbox_inches="tight")
            plt.close()

            # Sample composition by cluster (heatmap)
            fig, ax = plt.subplots(figsize=(12, 8))
            sns.heatmap(cluster_sample_props.T, annot=True, fmt=".2f", cmap="Blues", ax=ax, cbar_kws={"label": "Proportion"})
            ax.set_title(f"Sample Distribution Across Clusters - {self.integration_method.upper()}")
            ax.set_xlabel("Cluster")
            ax.set_ylabel("Sample")

            plt.tight_layout()
            plt.savefig(figures_dir / f"sample_distribution_{self.integration_method}.png", dpi=300, bbox_inches="tight")
            plt.close()

        except Exception as e:
            logger.warning(f"Could not generate cluster composition plots: {e}")
            import traceback

            logger.debug(f"Traceback: {traceback.format_exc()}")

    def _generate_comprehensive_summary_plots(self, adata: sc.AnnData, figures_dir: Path, logger: logging.Logger):
        """Generate comprehensive summary and statistics plots."""
        try:
            logger.info("Generating comprehensive summary plots...")

            # Create a large summary figure
            fig = plt.figure(figsize=(20, 16))
            gs = fig.add_gridspec(4, 4, height_ratios=[1, 1, 1, 1], width_ratios=[1, 1, 1, 1])

            # Plot 1: Sample overview pie chart
            ax1 = fig.add_subplot(gs[0, 0])
            sample_counts = adata.obs[self.batch_key].value_counts()
            colors = plt.cm.Set3(np.linspace(0, 1, len(sample_counts)))
            wedges, texts, autotexts = ax1.pie(sample_counts.values, labels=sample_counts.index, autopct="%1.1f%%", colors=colors, startangle=90)
            ax1.set_title("Sample Distribution")

            # Plot 2: Cluster size distribution
            ax2 = fig.add_subplot(gs[0, 1])
            leiden_key = f"leiden_{self.integration_method}"
            if leiden_key in adata.obs.columns:
                cluster_counts = adata.obs[leiden_key].value_counts().sort_index()
                bars = ax2.bar(range(len(cluster_counts)), cluster_counts.values, color="skyblue", alpha=0.7)
                ax2.set_title("Cluster Size Distribution")
                ax2.set_xlabel("Cluster")
                ax2.set_ylabel("Number of Cells")
                ax2.tick_params(axis="x", rotation=45)

                # Add stats text
                ax2.text(
                    0.02,
                    0.98,
                    f"Total clusters: {len(cluster_counts)}\nLargest: {cluster_counts.max()}\nSmallest: {cluster_counts.min()}",
                    transform=ax2.transAxes,
                    va="top",
                    ha="left",
                    fontsize=8,
                    bbox=dict(boxstyle="round", facecolor="white", alpha=0.8),
                )
            else:
                ax2.text(0.5, 0.5, "No cluster information", ha="center", va="center", transform=ax2.transAxes)
                ax2.set_title("Cluster Distribution (N/A)")

            # Plot 3: Integration embedding variance
            ax3 = fig.add_subplot(gs[0, 2])
            integration_key = "X_scvi" if self.integration_method == "scvi" else "X_harmony"
            if integration_key in adata.obsm:
                embed_data = adata.obsm[integration_key]
                # Calculate explained variance for each dimension
                var_per_dim = np.var(embed_data, axis=0)
                var_prop = var_per_dim / np.sum(var_per_dim) * 100

                dims_to_show = min(10, len(var_prop))  # Show first 10 dimensions
                bars = ax3.bar(range(dims_to_show), var_prop[:dims_to_show], color="orange", alpha=0.7)
                ax3.set_title(f"{self.integration_method.upper()} Embedding Variance")
                ax3.set_xlabel("Dimension")
                ax3.set_ylabel("Variance Explained (%)")
                ax3.tick_params(axis="x", rotation=45)
            else:
                ax3.text(0.5, 0.5, "No embedding data", ha="center", va="center", transform=ax3.transAxes)
                ax3.set_title("Embedding Variance (N/A)")

            # Plot 4: Integration summary statistics
            ax4 = fig.add_subplot(gs[0, 3])
            ax4.axis("off")
            stats_text = f"""Integration Summary

Method: {self.integration_method.upper()}
Total Cells: {adata.n_obs:,}
Total Genes: {adata.n_vars:,}
Samples: {len(adata.obs[self.batch_key].unique())}

Sample Details:"""

            for sample, count in sample_counts.items():
                stats_text += f"\n  {sample}: {count:,} cells"

            if leiden_key in adata.obs.columns:
                stats_text += f"\n\nClusters: {len(adata.obs[leiden_key].unique())}"

            if "cell_type" in adata.obs.columns:
                known_types = adata.obs["cell_type"].ne("Unknown").sum()
                stats_text += f"\n\nCell Types Mapped: {known_types:,}/{adata.n_obs:,}"
                stats_text += f" ({known_types / adata.n_obs * 100:.1f}%)"

            ax4.text(0.05, 0.95, stats_text, transform=ax4.transAxes, va="top", ha="left", fontsize=10, family="monospace", bbox=dict(boxstyle="round", facecolor="lightgray", alpha=0.8))

            # Plot 5-8: Sample-wise UMAP subplots
            samples = adata.obs[self.batch_key].unique()
            for i, sample in enumerate(samples[:4]):  # Show first 4 samples
                row = 1 + i // 2
                col = i % 2
                ax = fig.add_subplot(gs[row, col])

                sample_mask = adata.obs[self.batch_key] == sample
                sample_data = adata[sample_mask]

                if leiden_key in adata.obs.columns:
                    cluster_colors = sample_data.obs[leiden_key].astype("category").cat.codes
                    scatter = ax.scatter(sample_data.obsm["X_umap"][:, 0], sample_data.obsm["X_umap"][:, 1], c=cluster_colors, s=0.5, alpha=0.7, cmap="tab20")
                else:
                    scatter = ax.scatter(sample_data.obsm["X_umap"][:, 0], sample_data.obsm["X_umap"][:, 1], s=0.5, alpha=0.7, color="blue")

                ax.set_title(f"{sample} ({sample_data.n_obs:,} cells)")
                ax.set_xlabel("UMAP 1")
                ax.set_ylabel("UMAP 2")
                ax.grid(True, alpha=0.3)

            # Plot 9: Batch mixing metrics visualization
            if len(samples) > 1:
                ax9 = fig.add_subplot(gs[2, 2:])

                # Calculate nearest neighbor batch mixing
                from sklearn.neighbors import NearestNeighbors

                nn = NearestNeighbors(n_neighbors=min(50, adata.n_obs // 10))
                nn.fit(adata.obsm["X_umap"])

                batch_mixing_scores = []
                for i in range(min(1000, adata.n_obs)):  # Sample subset for efficiency
                    distances, indices = nn.kneighbors(adata.obsm["X_umap"][i : i + 1])
                    neighbor_batches = adata.obs[self.batch_key].iloc[indices[0]]
                    original_batch = adata.obs[self.batch_key].iloc[i]

                    # Calculate proportion of neighbors from different batches
                    different_batch_prop = (neighbor_batches != original_batch).mean()
                    batch_mixing_scores.append(different_batch_prop)

                # Create histogram of mixing scores
                ax9.hist(batch_mixing_scores, bins=30, alpha=0.7, color="green", edgecolor="black")
                ax9.set_title("Batch Mixing Quality\n(Higher = Better Integration)")
                ax9.set_xlabel("Proportion of Neighbors from Different Batches")
                ax9.set_ylabel("Number of Cells")
                ax9.axvline(np.mean(batch_mixing_scores), color="red", linestyle="--", label=f"Mean: {np.mean(batch_mixing_scores):.3f}")
                ax9.legend()
                ax9.grid(True, alpha=0.3)

            plt.tight_layout()
            plt.savefig(figures_dir / f"integration_comprehensive_summary_{self.integration_method}.png", dpi=300, bbox_inches="tight")
            plt.close()

            logger.info("Comprehensive summary plots generated successfully")

        except Exception as e:
            logger.warning(f"Could not generate comprehensive summary plots: {e}")
            import traceback

            logger.debug(f"Traceback: {traceback.format_exc()}")

    def _generate_celltype_analysis_plots(self, adata: sc.AnnData, figures_dir: Path, logger: logging.Logger):
        """Generate cell type analysis plots."""
        try:
            logger.info("Generating cell type analysis plots...")

            # Get cell types that are not 'Unknown'
            known_celltypes = adata.obs[adata.obs["cell_type"] != "Unknown"]["cell_type"]
            if len(known_celltypes) == 0:
                logger.info("No known cell types found, skipping cell type analysis")
                return

            fig, axes = plt.subplots(2, 2, figsize=(20, 12))
            fig.suptitle("Cell Type Analysis", fontsize=16)

            # Plot 1: Cell type distribution
            ax1 = axes[0, 0]
            celltype_counts = known_celltypes.value_counts()
            colors = plt.cm.Set3(np.linspace(0, 1, len(celltype_counts)))
            bars = ax1.bar(range(len(celltype_counts)), celltype_counts.values, color=colors, alpha=0.8)
            ax1.set_title("Cell Type Distribution")
            ax1.set_xlabel("Cell Type")
            ax1.set_ylabel("Number of Cells")
            ax1.set_xticks(range(len(celltype_counts)))
            ax1.set_xticklabels(celltype_counts.index, rotation=45, ha="right")
            ax1.grid(True, alpha=0.3, axis="y")

            # Add count labels on bars
            for i, bar in enumerate(bars):
                ax1.text(bar.get_x() + bar.get_width() / 2, bar.get_height() + max(celltype_counts) * 0.01, f"{celltype_counts.values[i]}", ha="center", va="bottom", fontsize=8)

            # Plot 2: Cell type by sample heatmap
            ax2 = axes[0, 1]
            if len(adata.obs[self.batch_key].unique()) > 1:
                celltype_sample_crosstab = pd.crosstab(adata.obs["cell_type"], adata.obs[self.batch_key])
                # Only include known cell types
                celltype_sample_crosstab = celltype_sample_crosstab.loc[celltype_sample_crosstab.index != "Unknown"]

                if len(celltype_sample_crosstab) > 0:
                    sns.heatmap(celltype_sample_crosstab, annot=True, fmt="d", cmap="Blues", ax=ax2, cbar_kws={"label": "Cell Count"})
                    ax2.set_title("Cell Types by Sample")
                    ax2.set_xlabel("Sample")
                    ax2.set_ylabel("Cell Type")
                else:
                    ax2.text(0.5, 0.5, "No cell type data", ha="center", va="center", transform=ax2.transAxes)
                    ax2.set_title("Cell Types by Sample (N/A)")
            else:
                ax2.text(0.5, 0.5, "Single sample data", ha="center", va="center", transform=ax2.transAxes)
                ax2.set_title("Cell Types by Sample (Single Sample)")

            # Plot 3: Cell type cluster enrichment
            ax3 = axes[1, 0]
            leiden_key = f"leiden_{self.integration_method}"
            if leiden_key in adata.obs.columns:
                # Calculate cell type enrichment per cluster
                cluster_celltype_crosstab = pd.crosstab(adata.obs[leiden_key], adata.obs["cell_type"])
                # Normalize by cluster size to get proportions
                cluster_celltype_props = cluster_celltype_crosstab.div(cluster_celltype_crosstab.sum(axis=1), axis=0)
                # Remove 'Unknown' column if it exists
                if "Unknown" in cluster_celltype_props.columns:
                    cluster_celltype_props = cluster_celltype_props.drop("Unknown", axis=1)

                if len(cluster_celltype_props.columns) > 0:
                    sns.heatmap(cluster_celltype_props.T, annot=True, fmt=".1f", cmap="Reds", ax=ax3, cbar_kws={"label": "Proportion"})
                    ax3.set_title("Cell Type Enrichment per Cluster")
                    ax3.set_xlabel("Cluster")
                    ax3.set_ylabel("Cell Type")
                else:
                    ax3.text(0.5, 0.5, "No enrichment data", ha="center", va="center", transform=ax3.transAxes)
                    ax3.set_title("Cell Type Enrichment (N/A)")
            else:
                ax3.text(0.5, 0.5, "No cluster data", ha="center", va="center", transform=ax3.transAxes)
                ax3.set_title("Cell Type Enrichment (N/A)")

            # Plot 4: Cell type statistics
            ax4 = axes[1, 1]
            ax4.axis("off")

            stats_text = "Cell Type Statistics\n\n"
            stats_text += f"Total cells with known types: {len(known_celltypes):,}\n"
            stats_text += f"Total cells: {adata.n_obs:,}\n"
            stats_text += f"Coverage: {len(known_celltypes) / adata.n_obs * 100:.1f}%\n\n"
            stats_text += f"Cell types identified: {len(celltype_counts)}\n\n"

            stats_text += "Top cell types:\n"
            for celltype, count in celltype_counts.head(8).items():
                percentage = count / len(known_celltypes) * 100
                stats_text += f"  {celltype}: {count:,} ({percentage:.1f}%)\n"

            ax4.text(0.05, 0.95, stats_text, transform=ax4.transAxes, va="top", ha="left", fontsize=10, family="monospace", bbox=dict(boxstyle="round", facecolor="lightblue", alpha=0.8))

            plt.tight_layout()
            plt.savefig(figures_dir / f"celltype_analysis_{self.integration_method}.png", dpi=300, bbox_inches="tight")
            plt.close()

            logger.info("Cell type analysis plots generated successfully")

        except Exception as e:
            logger.warning(f"Could not generate cell type analysis plots: {e}")
            import traceback

            logger.debug(f"Traceback: {traceback.format_exc()}")

    def _generate_scib_metrics(self, adata: sc.AnnData, figures_dir: Path, logger: logging.Logger):
        """Generate scib metrics for integration evaluation."""
        try:
            logger.info("Generating scib integration metrics...")

            # Check for scib availability
            try:
                from scib_metrics.benchmark import BatchCorrection, Benchmarker, BioConservation
            except ImportError:
                logger.warning("scib-metrics package not available. Install with: pip install scib-metrics")
                console.print("[yellow]⚠ scib-metrics package not available for metrics evaluation[/yellow]")
                return

            # Prepare data for scib evaluation
            adata_scib = adata.copy()

            # Filter out Unknown cell types for evaluation
            known_mask = adata_scib.obs["cell_type"] != "Unknown"
            adata_scib = adata_scib[known_mask].copy()

            if len(adata_scib) == 0:
                logger.warning("No cells with known cell types for scib evaluation")
                return

            logger.info(f"Evaluating integration on {adata_scib.n_obs} cells with known cell types")

            # Determine available embeddings
            available_embeddings = []
            integration_key = "X_scvi" if self.integration_method == "scvi" else "X_harmony"

            # Add unintegrated representation (PCA if available, otherwise raw X)
            if "X_pca" in adata_scib.obsm:
                available_embeddings.append("X_pca")
                # Rename for scib
                adata_scib.obsm["Unintegrated"] = adata_scib.obsm["X_pca"].copy()
            else:
                # Use raw expression as unintegrated
                adata_scib.obsm["Unintegrated"] = adata_scib.X.toarray() if hasattr(adata_scib.X, "toarray") else adata_scib.X
                available_embeddings.append("Unintegrated")

            # Add integrated representation
            if integration_key in adata_scib.obsm:
                method_name = self.integration_method.upper()
                adata_scib.obsm[method_name] = adata_scib.obsm[integration_key].copy()
                available_embeddings.append(method_name)
            else:
                logger.warning(f"Integration embedding {integration_key} not found in obsm")
                return

            logger.info(f"Available embeddings for evaluation: {available_embeddings}")

            # Create benchmarker using scib-metrics
            bm = Benchmarker(
                adata_scib,
                batch_key=self.batch_key,
                label_key="cell_type",
                bio_conservation_metrics=BioConservation(),
                batch_correction_metrics=BatchCorrection(),
                embedding_obsm_keys=available_embeddings,
                n_jobs=1,
            )

            # Run benchmarking
            logger.info("Running scib benchmarking...")
            console.print("[dim]Running scib integration benchmarking (this may take a while)...[/dim]")

            bm.benchmark()
            results_df = bm.get_results(min_max_scale=False)

            # Save results table
            results_path = figures_dir / f"scib_metrics_{self.integration_method}.csv"
            results_df.to_csv(results_path)
            logger.info(f"scib metrics saved to {results_path}")

            # Create visualization of results
            fig, axes = plt.subplots(2, 2, figsize=(16, 12))
            fig.suptitle(f"scib Integration Metrics - {self.integration_method.upper()}", fontsize=16)

            # Get metrics categories
            bio_metrics = [col for col in results_df.columns if any(bio_term in col.lower() for bio_term in ["nmi", "ari", "silhouette", "isolated", "trajectory", "hvg"])]
            batch_metrics = [col for col in results_df.columns if any(batch_term in col.lower() for batch_term in ["pcr", "batch", "kbet", "graph", "asw"])]

            # Plot 1: Biological conservation metrics
            ax1 = axes[0, 0]
            if bio_metrics and len(results_df) > 1:
                bio_data = results_df[bio_metrics].T
                sns.heatmap(bio_data, annot=True, fmt=".3f", cmap="Greens", ax=ax1, cbar_kws={"label": "Score"})
                ax1.set_title("Biological Conservation Metrics")
                ax1.set_xlabel("Method")
                ax1.set_ylabel("Metric")
            else:
                ax1.text(0.5, 0.5, "Insufficient data\\nfor bio metrics", ha="center", va="center", transform=ax1.transAxes)
                ax1.set_title("Biological Conservation Metrics (N/A)")

            # Plot 2: Batch correction metrics
            ax2 = axes[0, 1]
            if batch_metrics and len(results_df) > 1:
                batch_data = results_df[batch_metrics].T
                sns.heatmap(batch_data, annot=True, fmt=".3f", cmap="Blues", ax=ax2, cbar_kws={"label": "Score"})
                ax2.set_title("Batch Correction Metrics")
                ax2.set_xlabel("Method")
                ax2.set_ylabel("Metric")
            else:
                ax2.text(0.5, 0.5, "Insufficient data\\nfor batch metrics", ha="center", va="center", transform=ax2.transAxes)
                ax2.set_title("Batch Correction Metrics (N/A)")

            # Plot 3: Overall metrics comparison (if multiple methods)
            ax3 = axes[1, 0]
            if len(results_df) > 1:
                # Select key metrics for comparison
                key_metrics = []
                for metric in ["NMI_cluster/label", "ARI_cluster/label", "ASW_label_batch", "PCR_batch"]:
                    if metric in results_df.columns:
                        key_metrics.append(metric)

                if key_metrics:
                    key_data = results_df[key_metrics]
                    x_pos = np.arange(len(key_metrics))
                    width = 0.35

                    for i, method in enumerate(results_df.index):
                        offset = (i - len(results_df.index) / 2 + 0.5) * width / len(results_df.index)
                        bars = ax3.bar(x_pos + offset, key_data.loc[method], width / len(results_df.index), label=method, alpha=0.7)

                        # Add value labels on bars
                        for j, bar in enumerate(bars):
                            ax3.text(bar.get_x() + bar.get_width() / 2, bar.get_height() + 0.01, f"{key_data.loc[method].iloc[j]:.3f}", ha="center", va="bottom", fontsize=8)

                    ax3.set_xlabel("Metrics")
                    ax3.set_ylabel("Score")
                    ax3.set_title("Key Metrics Comparison")
                    ax3.set_xticks(x_pos)
                    ax3.set_xticklabels(key_metrics, rotation=45, ha="right")
                    ax3.legend()
                    ax3.grid(True, alpha=0.3, axis="y")
                else:
                    ax3.text(0.5, 0.5, "No key metrics found", ha="center", va="center", transform=ax3.transAxes)
                    ax3.set_title("Key Metrics (N/A)")
            else:
                ax3.text(0.5, 0.5, "Single method\\nNo comparison", ha="center", va="center", transform=ax3.transAxes)
                ax3.set_title("Metrics Comparison (N/A)")

            # Plot 4: Metrics summary table
            ax4 = axes[1, 1]
            ax4.axis("off")

            # Create summary text
            summary_text = "scib Metrics Summary\\n"
            summary_text += f"{'=' * 30}\\n\\n"
            summary_text += f"Method: {self.integration_method.upper()}\\n"
            summary_text += f"Cells evaluated: {adata_scib.n_obs:,}\\n"
            summary_text += f"Batches: {len(adata_scib.obs[self.batch_key].unique())}\\n"
            summary_text += f"Cell types: {len(adata_scib.obs['cell_type'].unique())}\\n\\n"

            # Add key metric values
            if len(results_df) > 0:
                method_name = results_df.index[0] if len(results_df) == 1 else self.integration_method.upper()
                if method_name in results_df.index:
                    summary_text += f"Key Metrics for {method_name}:\\n"
                    for metric in ["NMI_cluster/label", "ARI_cluster/label", "ASW_label_batch", "PCR_batch"]:
                        if metric in results_df.columns:
                            value = results_df.loc[method_name, metric]
                            summary_text += f"  {metric}: {value:.3f}\\n"

                    summary_text += "\\nMetric Interpretation:\\n"
                    summary_text += "• NMI/ARI: Bio conservation (higher=better)\\n"
                    summary_text += "• ASW_label_batch: Batch correction (lower=better)\\n"
                    summary_text += "• PCR_batch: Batch effect (lower=better)\\n"

            ax4.text(0.05, 0.95, summary_text, transform=ax4.transAxes, va="top", ha="left", fontsize=10, family="monospace", bbox=dict(boxstyle="round", facecolor="lightgray", alpha=0.8))

            plt.tight_layout()
            plt.savefig(figures_dir / f"scib_metrics_{self.integration_method}.png", dpi=300, bbox_inches="tight")
            plt.close()

            logger.info("scib metrics evaluation completed successfully")
            console.print(f"[green]✓ scib metrics generated and saved to {results_path}[/green]")

        except Exception as e:
            logger.warning(f"Could not generate scib metrics: {e}")
            import traceback

            logger.debug(f"Traceback: {traceback.format_exc()}")
            console.print(f"[yellow]⚠ Warning: scib metrics evaluation failed: {e}[/yellow]")

    def add_cli_args(self, parser: argparse.ArgumentParser) -> None:
        """Add CLI arguments for the Integration function."""
        parser.add_argument("--method", "-m", choices=["scvi", "harmony"], default="harmony", help="Integration method to use (default: harmony)")
        parser.add_argument("--output-name", "-o", type=str, help="Output file name (without extension)")
        parser.add_argument("--sample", "-s", type=str, help="Comma-separated list of specific sample IDs to integrate (e.g., GSM3934448,GSM3934449)")
        parser.add_argument("--n-pcs", type=int, default=50, help="Number of principal components to use (default: 50)")
        parser.add_argument("--batch-key", type=str, default="sample", help="Column name in obs containing batch information (default: sample)")
        parser.add_argument("--scvi-epochs", type=int, default=200, help="Number of training epochs for scVI (default: 200)")
        parser.add_argument("--no-early-stopping", action="store_true", help="Disable early stopping in scVI training")
        parser.add_argument("--harmony-vars", nargs="+", help="Variables to use for Harmony correction (default: uses batch-key)")
        parser.add_argument("--force-rerun", action="store_true", help="Force rerun even if cached results exist")
        parser.add_argument("--quiet", action="store_true", help="Reduce verbosity")

    def cli(self, project, args: argparse.Namespace | None = None):
        """CLI entry point for Integration function."""
        # Extract arguments with defaults
        integration_method = "harmony"
        output_name = None
        n_pcs = 50
        batch_key = "sample"
        scvi_epochs = 200
        scvi_early_stopping = True
        harmony_vars_use = None
        force_rerun = False
        verbose = True
        sample_list = None

        if args:
            if hasattr(args, "method"):
                integration_method = args.method
            if hasattr(args, "output_name"):
                output_name = args.output_name
            if hasattr(args, "n_pcs"):
                n_pcs = args.n_pcs
            if hasattr(args, "batch_key"):
                batch_key = args.batch_key
            if hasattr(args, "scvi_epochs"):
                scvi_epochs = args.scvi_epochs
            if hasattr(args, "no_early_stopping"):
                scvi_early_stopping = not args.no_early_stopping
            if hasattr(args, "harmony_vars"):
                harmony_vars_use = args.harmony_vars
            if hasattr(args, "force_rerun"):
                force_rerun = args.force_rerun
            if hasattr(args, "quiet"):
                verbose = not args.quiet
            if hasattr(args, "sample") and args.sample:
                sample_list = [s.strip() for s in args.sample.split(",")]

        console.print(f"[cyan]Starting integration with {integration_method.upper()}...[/cyan]")
        if sample_list:
            console.print(f"Target samples: {', '.join(sample_list)}")
        console.print(f"Batch key: {batch_key}")
        console.print(f"Number of PCs: {n_pcs}")
        if integration_method == "scvi":
            console.print(f"scVI epochs: {scvi_epochs}")
            console.print(f"Early stopping: {scvi_early_stopping}")
        elif integration_method == "harmony":
            vars_display = harmony_vars_use if harmony_vars_use else [batch_key]
            console.print(f"Harmony variables: {', '.join(vars_display)}")

        integrate_instance = Integrate(
            integration_method=integration_method,
            outfile_name=output_name,
            n_pcs=n_pcs,
            batch_key=batch_key,
            scvi_epochs=scvi_epochs,
            scvi_early_stopping=scvi_early_stopping,
            harmony_vars_use=harmony_vars_use,
            force_rerun=force_rerun,
            verbose=verbose,
            sample_list=sample_list,
        )
        return integrate_instance.call(project)

    def get_description(self) -> str:
        return """Integrate single-cell data using scVI or Harmony methods.

This function performs batch correction and data integration using either:
- scVI: Deep generative model for probabilistic integration
- Harmony: Fast linear method for batch correction using fuzzy clustering

Both methods support multi-sample integration and produce UMAP embeddings for visualization."""

    def get_usage_examples(self) -> list[str]:
        return [
            "celline run integrate",
            "celline run integrate --method scvi --scvi-epochs 300",
            "celline run integrate --method harmony --harmony-vars sample project",
            "celline run integrate --output-name my_integration --n-pcs 30",
            "celline run integrate --force-rerun --quiet",
        ]



from __future__ import annotations

import argparse
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from celline import Project


class CellineFunction(ABC):
    """Abstract base class for celline functions"""

    def __init__(self, **kwargs) -> None:
        super().__init__()

    def register(self) -> str:
        """Register method - return method name for command"""
        return self.__class__.__name__.lower()

    @abstractmethod
    def call(self, project: "Project"):
        """Abstract method to be implemented by subclasses"""
        pass

    def cli(self, project: "Project", args: Optional[argparse.Namespace] = None) -> "Project":
        """
        CLI entry point for this function.
        Override this method to provide CLI-specific functionality.
        
        Args:
            project: The Project instance
            args: Parsed CLI arguments (if any)
            
        Returns:
            Project: The project instance (for chaining)
        """
        # Default implementation calls the regular call method
        return self.call(project)

    def add_cli_args(self, parser: argparse.ArgumentParser) -> None:
        """
        Add CLI arguments specific to this function.
        Override this method to add custom arguments.
        
        Args:
            parser: The ArgumentParser to add arguments to
        """
        pass

    def get_description(self) -> str:
        """
        Get a description of this function for CLI help.
        Override this method to provide a better description.
        
        Returns:
            str: Description of the function
        """
        return self.__doc__ or f"{self.__class__.__name__} function"

    def get_usage_examples(self) -> list[str]:
        """
        Get usage examples for CLI help.
        Override this method to provide usage examples.
        
        Returns:
            list[str]: List of usage example strings
        """
        return []



from typing import NamedTuple
from typing import Dict, Any, get_type_hints, TypeVar, Generic, List, Type
import inspect
import polars as pl


class NamedTupleAndDictStructure:
    @staticmethod
    def serialize(obj) -> Dict[str, Any]:
        """Serialize given obj (inherits NamedTuple) to dictionary type"""
        if isinstance(obj, tuple) and hasattr(
            obj, "_fields"
        ):  # check if obj is a namedtuple or subclass
            return {f: NamedTupleAndDictStructure.serialize(getattr(obj, f)) for f in obj._fields}  # type: ignore
        elif isinstance(obj, list):
            return [NamedTupleAndDictStructure.serialize(v) for v in obj]  # type: ignore
        else:
            return obj  # type: ignore

    T = TypeVar("T")

    @staticmethod
    def deserialize(data: Dict, namedtuple_cls: T) -> T:
        """Deserialize give dictionary to give type T instance"""
        if (
            inspect.isclass(namedtuple_cls)
            and issubclass(namedtuple_cls, tuple)
            and hasattr(namedtuple_cls, "_fields")
        ):
            fields = {
                f: NamedTupleAndDictStructure.deserialize(data.get(f, None), t)
                for f, t in get_type_hints(namedtuple_cls).items()
            }
            return namedtuple_cls(**fields)  # type: ignore
        elif isinstance(namedtuple_cls, list) and len(namedtuple_cls) == 1:
            return [NamedTupleAndDictStructure.deserialize(v, namedtuple_cls[0]) for v in data]  # type: ignore
        else:
            return data  # type: ignore


T = TypeVar("T", bound=NamedTuple)


class NamedTupleAndPolarsStructure(Generic[T]):
    @staticmethod
    def serialize(obj: NamedTuple) -> pl.DataFrame:
        return pl.DataFrame(
            {field: [getattr(obj, field)] for field in obj._fields},
            schema=get_type_hints(obj),
        )

    @staticmethod
    def deserialize(df: pl.DataFrame, tuple_type: Type[T]) -> List[T]:
        return [tuple_type(*t) for t in df.to_pandas().itertuples(index=False)]  # type: ignore



import argparse
from typing import Optional
from rich.console import Console

from celline.functions._base import CellineFunction

console = Console()


class Info(CellineFunction):
    def register(self) -> str:
        return "info"

    def call(self, project):
        console.print("[cyan]Project information:[/cyan]")
        console.print(f"Project path: {project.PROJ_PATH}")
        console.print(f"Exec path: {project.EXEC_PATH}")
        return project

    def cli(self, project, args: Optional[argparse.Namespace] = None):
        """CLI entry point for Info function."""
        return self.call(project)

    def get_description(self) -> str:
        return "Display project information and status."

    def get_usage_examples(self) -> list[str]:
        return ["celline run info"]



# from __future__ import annotations
# import os
# import glob

# from typing import Dict, NamedTuple, Final, List

# import toml

# from celline.config import Config
# from celline.utils.path import Path
# from celline.utils.serialization import NamedTupleAndPolarsStructure
# from celline.DB.model import SRA_GSM


# class Resources:
#     class SampleInfo:
#         """#### Represents sample information"""

#         def __init__(self, sample_id: str, sample_name: str) -> None:
#             self.sample_id = sample_id
#             project_id = SRA_GSM().search(sample_id).parent
#             if project_id is None:
#                 raise KeyError(f"Parent ID of {project_id} is null")
#             self.project_id = project_id
#             self._name = sample_name
#             self.path = Path(project_id, sample_id)

#         sample_id: Final[str]
#         """[Immutable] Sample ID like GSM..."""
#         project_id: Final[str]
#         """[Immutable] Project ID like GSE"""
#         _name: str
#         """Sample name"""
#         path: Final[Path]
#         """[Immutable] Path object which enable search path"""

#         @property
#         def downloaded(self):
#             """Already downloaded?"""
#             # resources_sample_raw_fastqs以下の全てのディレクトリを取得
#             all_dirs = [
#                 f.path
#                 for f in os.scandir(self.path.resources_sample_raw_fastqs)
#                 if f.is_dir()
#             ]
#             # 各ディレクトリに<任意の文字列>_S1_L001_<R1, R2, I1, I2のいずれか>_001.fastq.gzという名前のファイルが最低でも2つ含まれているか
#             for directory in all_dirs:
#                 fastq_files = []
#                 for suffix in ["R1", "R2", "I1", "I2"]:  # R1, R2, I1, I2の各パターンについて
#                     fastq_files += glob.glob(
#                         f"{directory}/*_S1_L001_{suffix}_001.fastq.gz"
#                     )  # 各パターンにマッチするファイルを取得

#                 if len(fastq_files) >= 2:
#                     return True  # 該当のファイルが2つ以上含まれるディレクトリが見つかった

#             return False

#         @property
#         def counted(self):
#             """Already counted?"""
#             counted_file = f"{self.path.resources_sample_counted}/outs/filtered_feature_bc_matrix.h5"
#             return os.path.isfile(counted_file)

#         @property
#         def preprocessed(self) -> bool:
#             return os.path.isfile(
#                 f"{self.path.data_sample}/doublet_info.tsv"
#             ) and os.path.isfile(f"{self.path.data_sample}/qc_matrix.tsv")

#         @property
#         def celltype_predicted(self):
#             return os.path.isfile(f"{self.path.data_sample}/celltype_predicted.tsv")

#         @property
#         def name(self):
#             return self._name

#         @name.setter
#         def name(self, new_name: str):
#             self._name = new_name
#             all_samples = Resources.all_samples()
#             for i, sample in enumerate(all_samples):
#                 if sample.sample_id == self.sample_id:
#                     all_samples[i].name = new_name
#             Resources.save_samples(all_samples)

#     @classmethod
#     def all_samples(cls) -> List[SampleInfo]:
#         sample_info_file = f"{Config.PROJ_ROOT}/samples.toml"
#         samples: Dict[str, str] = {}
#         if os.path.isfile(sample_info_file):
#             with open(sample_info_file, mode="r", encoding="utf-8") as f:
#                 samples = toml.load(f)
#             return [Resources.SampleInfo(sample, samples[sample]) for sample in samples]
#         return []

#     @classmethod
#     def save_samples(cls, samples: List[SampleInfo]) -> None:
#         sample_info_file = f"{Config.PROJ_ROOT}/samples.toml"
#         samples_dict = {sample.sample_id: sample.name for sample in samples}
#         with open(sample_info_file, mode="w", encoding="utf-8") as f:
#             toml.dump(samples_dict, f)



from __future__ import annotations

from enum import Enum
from typing import Optional

from celline.decorators import classproperty


class ServerSystem:
    class JobType(Enum):
        """An enumeration that specifies the type of job system to use."""

        MultiThreading = 1
        PBS = 2

    job_system: JobType = JobType.MultiThreading
    cluster_server_name: str | None = None

    @classmethod
    def useMultiThreading(cls):
        cls.job_system = ServerSystem.JobType.MultiThreading
        cls.cluster_server_name = None

    @classmethod
    def usePBS(cls, cluster_server_name: str):
        if cluster_server_name is None or cluster_server_name.strip() == "":
            raise ValueError("cluster_server_name cannot be None or empty")
        cls.job_system = ServerSystem.JobType.PBS
        cls.cluster_server_name = cluster_server_name.strip()

    @classmethod
    def is_pbs_configured(cls) -> bool:
        """Check if PBS is properly configured."""
        return cls.job_system == ServerSystem.JobType.PBS and cls.cluster_server_name is not None and cls.cluster_server_name.strip() != ""

    @classmethod
    def get_status(cls) -> str:
        """Get current configuration status."""
        if cls.job_system == ServerSystem.JobType.PBS:
            if cls.is_pbs_configured():
                return f"PBS (cluster: {cls.cluster_server_name})"
            return "PBS (not configured - cluster_server_name is empty)"
        return "MultiThreading"



from celline.functions._base import CellineFunction


class Job(CellineFunction):
    def register(self) -> str:
        return "job"

    def call(self, project):
        print("Job management - implement functionality as needed")
        return project



from typing import Callable, List, Final, NamedTuple, Type, TypeVar
from celline.DB.dev.model import BaseModel, Primary, RunSchema


class CNCB_CRR_Schema(RunSchema):
    summary: str
    species: str


class CNCB_CRR(BaseModel):
    def set_class_name(self) -> str:
        return __class__.__name__

    def set_schema(self) -> type[CNCB_CRR_Schema]:
        return CNCB_CRR_Schema



"""
Function invoker for handling CellineFunction instantiation and execution.
"""

import inspect
import json
from typing import Any, Dict, List, Optional, Type
from rich.console import Console
from rich.prompt import Prompt, Confirm

from celline.functions._base import CellineFunction
from celline.interfaces import Project

console = Console()


class FunctionInvoker:
    """Handles instantiation and execution of CellineFunction classes."""
    
    def __init__(self, func_class: Type[CellineFunction]):
        self.func_class = func_class
        self.signature = inspect.signature(func_class.__init__)
    
    def get_constructor_params(self) -> Dict[str, inspect.Parameter]:
        """Get constructor parameters excluding 'self'."""
        params = {}
        for name, param in self.signature.parameters.items():
            if name != 'self':
                params[name] = param
        return params
    
    def can_instantiate_without_args(self) -> bool:
        """Check if the function can be instantiated without arguments."""
        params = self.get_constructor_params()
        for param in params.values():
            if param.default == inspect.Parameter.empty:
                return False
        return True
    
    def prompt_for_args(self) -> Dict[str, Any]:
        """Interactively prompt user for constructor arguments."""
        params = self.get_constructor_params()
        args = {}
        
        console.print(f"[yellow]Function '{self.func_class.__name__}' requires arguments:[/yellow]")
        console.print()
        
        for name, param in params.items():
            param_type = param.annotation
            default_value = param.default if param.default != inspect.Parameter.empty else None
            
            # Format the prompt
            prompt_text = f"{name}"
            if param_type != inspect.Parameter.empty:
                prompt_text += f" ({param_type.__name__ if hasattr(param_type, '__name__') else str(param_type)})"
            if default_value is not None:
                prompt_text += f" [default: {default_value}]"
            
            # Get user input
            if param_type == bool or str(param_type) == "<class 'bool'>":
                value = Confirm.ask(prompt_text, default=default_value if default_value is not None else False)
            else:
                value = Prompt.ask(prompt_text, default=str(default_value) if default_value is not None else None)
                
                # Try to convert to appropriate type
                if value and param_type != inspect.Parameter.empty:
                    try:
                        if param_type == int:
                            value = int(value)
                        elif param_type == float:
                            value = float(value)
                        elif param_type == bool:
                            value = value.lower() in ('true', 'yes', '1', 'on')
                        elif hasattr(param_type, '__origin__'):  # Handle generic types like List[str]
                            if str(param_type).startswith('typing.List') or str(param_type).startswith('list'):
                                # Simple list parsing - split by comma
                                value = [item.strip() for item in value.split(',') if item.strip()]
                    except (ValueError, TypeError) as e:
                        console.print(f"[yellow]Warning: Could not convert '{value}' to {param_type}, using as string[/yellow]")
            
            args[name] = value
        
        return args
    
    def create_instance(self, args: Optional[Dict[str, Any]] = None) -> CellineFunction:
        """Create an instance of the function with given arguments."""
        if args is None:
            args = {}
        
        try:
            return self.func_class(**args)
        except TypeError as e:
            raise ValueError(f"Could not instantiate {self.func_class.__name__}: {e}")
    
    def invoke(self, project: Project, interactive: bool = True) -> Any:
        """Invoke the function, prompting for args if needed."""
        try:
            # Try to create instance without arguments first
            if self.can_instantiate_without_args():
                instance = self.create_instance()
            elif interactive:
                # Prompt for arguments
                args = self.prompt_for_args()
                instance = self.create_instance(args)
            else:
                raise ValueError(f"Function {self.func_class.__name__} requires arguments but interactive mode is disabled")
            
            # Call the function
            console.print(f"[cyan]Executing {self.func_class.__name__}...[/cyan]")
            result = instance.call(project)
            console.print(f"[green]✓ {self.func_class.__name__} completed successfully[/green]")
            return result
            
        except KeyboardInterrupt:
            console.print("[yellow]Operation cancelled by user[/yellow]")
            raise
        except Exception as e:
            console.print(f"[red]Error: {e}[/red]")
            raise


def invoke_function(func_class: Type[CellineFunction], project: Project, interactive: bool = True) -> Any:
    """Convenience function to invoke a CellineFunction."""
    invoker = FunctionInvoker(func_class)
    return invoker.invoke(project, interactive)


import os
import subprocess
from typing import TYPE_CHECKING, Callable, Dict, Final, List, NamedTuple, Optional

import rich

from celline.config import Config, Setting
from celline.functions._base import CellineFunction
from celline.middleware import ThreadObservable
from celline.sample import SampleResolver
from celline.server import ServerSystem
from celline.template import TemplateManager

if TYPE_CHECKING:
    from celline import Project


class CreateSeuratObject(CellineFunction):
    class JobContainer(NamedTuple):
        """
        Represents job information for data download.
        """

        nthread: str
        cluster_server: str
        jobname: str
        logpath: str
        r_path: str
        exec_root: str
        input_h5_path: str
        data_dir_path: str
        proj_name: str
        useqc_matrix: str

    def __init__(
        self,
        useqc_matrix: bool,
        then: Optional[Callable[[str], None]] = None,
        catch: Optional[Callable[[subprocess.CalledProcessError], None]] = None,
    ) -> None:
        self.then: Final = then
        self.catch: Final = catch
        self.cluster_server: Final[Optional[str]] = ServerSystem.cluster_server_name
        self.useqc_matrix = "true" if useqc_matrix else "false"

    def call(self, project: "Project"):
        all_job_files: List[str] = []
        for sample in SampleResolver.samples.values():
            if not (sample.path.is_predicted_celltype and sample.path.is_preprocessed):
                rich.print(
                    f"[bold red]Sample {sample.schema.key} is not counted/predicted or pre-processed yet.[/]: Skip"
                )
            else:
                if os.path.isfile(f"{sample.path.data_sample}/seurat.rds"):
                    rich.print(
                        f":rocket:[bold green]Sample {sample.schema.key} is already processed.[/] Skip"
                    )
                else:
                    input_h5_path = f"{sample.path.resources_sample_counted}/outs/filtered_feature_bc_matrix.h5"
                    data_dir_path = sample.path.data_sample
                    TemplateManager.replace_from_file(
                        file_name="create_seurat.sh",
                        structure=CreateSeuratObject.JobContainer(
                            nthread=str(1),
                            cluster_server=""
                            if self.cluster_server is None
                            else self.cluster_server,
                            jobname="Create Seurat",
                            logpath=sample.path.data_log_file("create_seurat"),
                            r_path=f"{Setting.r_path}script",
                            exec_root=Config.EXEC_ROOT,
                            input_h5_path=input_h5_path,
                            data_dir_path=data_dir_path,
                            proj_name=Setting.name,
                            useqc_matrix=self.useqc_matrix,
                        ),
                        replaced_path=f"{sample.path.data_sample_src}/create_seurat.sh",
                    )
                    all_job_files.append(
                        f"{sample.path.data_sample_src}/create_seurat.sh"
                    )
        ThreadObservable.call_shell(all_job_files, proc_name="Creating seurat").watch()
        return project



from typing import Any, Generic, TypeVar, overload

from celline.plugins.reflection.type import TypeC, typeof

T = TypeVar("T")


class Activator:
    @staticmethod
    @overload
    def CreateInstance(t: typeof[T], *args: Any) -> T:
        ...

    @staticmethod
    @overload
    def CreateInstance(t: TypeC, *args: Any) -> Any:
        ...

    @staticmethod
    def CreateInstance(t: Any, *args: Any):
        if isinstance(t, typeof):
            if (len(args) == 0):
                return t.TypeInfo.BuiltinType()
            else:
                return t.TypeInfo.BuiltinType(args)
        if isinstance(t, TypeC):
            if (len(args) == 0):
                return t.BuiltinType()
            else:
                return t.BuiltinType(args)



import os
from typing import TYPE_CHECKING, Dict, Final, List, NamedTuple, Optional, Union

import polars as pl
import toml
import tqdm

from celline.DB.handler import GEOHandler
from celline.DB.model import SRA_GSE, SRA_GSM, SRA_SRR
from celline.DB.model.transcriptome import Transcriptome
from celline.config import Config
from celline.functions._base import CellineFunction
from celline.utils.serialization import NamedTupleAndPolarsStructure

if TYPE_CHECKING:
    from celline import Project


class SetTranscriptome(CellineFunction):
    def __init__(self, species: str, built_path: str, force_update=True) -> None:
        self.species: Final[str] = species
        self.built_path: Final[str] = built_path
        self.force_update: Final[bool] = force_update

    def call(self, project: "Project"):
        Transcriptome().add_path(
            self.species, self.built_path, force_update=self.force_update
        )
        return project



import subprocess
import shutil
import re
import os
from pathlib import Path
from typing import List, NamedTuple, Optional, Set
from rich.console import Console
from rich.table import Table
from rich.prompt import Confirm
import inquirer

console = Console()


class DependencyCheck(NamedTuple):
    """Represents a dependency check result."""
    name: str
    available: bool
    path: Optional[str] = None
    version: Optional[str] = None


class RPackageCheck(NamedTuple):
    """Represents an R package check result."""
    name: str
    available: bool
    version: Optional[str] = None


class DependencyValidator:
    """Validates system dependencies for Celline."""
    
    REQUIRED_DEPENDENCIES = [
        "cellranger",
        "R",
        "fastq-dump"
    ]
    
    OPTIONAL_DEPENDENCIES = [
        "wget",
        "curl"
    ]
    
    # Required R packages found in the codebase
    REQUIRED_R_PACKAGES = [
        "pacman",
        "Seurat", 
        "SeuratDisk",
        "SeuratObject",
        "tidyverse",
        "scPred",
        "doParallel",
        "scran",
        "batchelor",
        "Matrix"
    ]
    
    @staticmethod
    def check_command(command: str) -> DependencyCheck:
        """Check if a command is available in the system PATH."""
        path = shutil.which(command)
        if path:
            version = DependencyValidator._get_version(command)
            return DependencyCheck(name=command, available=True, path=path, version=version)
        else:
            return DependencyCheck(name=command, available=False)
    
    @staticmethod
    def _get_version(command: str) -> Optional[str]:
        """Get version of a command if possible."""
        try:
            if command == "cellranger":
                result = subprocess.run([command, "--version"], capture_output=True, text=True, timeout=10)
                if result.returncode == 0:
                    return result.stdout.strip().split('\n')[0]
            elif command == "R":
                result = subprocess.run([command, "--version"], capture_output=True, text=True, timeout=10)
                if result.returncode == 0:
                    lines = result.stdout.strip().split('\n')
                    for line in lines:
                        if "R version" in line:
                            return line.split()[2]
            elif command == "fastq-dump":
                result = subprocess.run([command, "--version"], capture_output=True, text=True, timeout=10)
                if result.returncode == 0:
                    return result.stdout.strip().split('\n')[0]
        except (subprocess.SubprocessError, subprocess.TimeoutExpired):
            pass
        return None
    
    @staticmethod
    def check_r_package(package: str, r_path: Optional[str] = None) -> RPackageCheck:
        """Check if an R package is installed."""
        r_command = r_path if r_path else "R"
        
        try:
            # Create R script to check package
            r_script = f'''
            if (require("{package}", quietly = TRUE, character.only = TRUE)) {{
                version <- packageVersion("{package}")
                cat("AVAILABLE:", version)
            }} else {{
                cat("NOT_AVAILABLE")
            }}
            '''
            
            result = subprocess.run(
                [r_command, "--slave", "--vanilla", "-e", r_script],
                capture_output=True, text=True, timeout=30
            )
            
            if result.returncode == 0:
                output = result.stdout.strip()
                if output.startswith("AVAILABLE:"):
                    version = output.replace("AVAILABLE:", "").strip()
                    return RPackageCheck(name=package, available=True, version=version)
                else:
                    return RPackageCheck(name=package, available=False)
            else:
                return RPackageCheck(name=package, available=False)
                
        except (subprocess.SubprocessError, subprocess.TimeoutExpired):
            return RPackageCheck(name=package, available=False)
    
    @staticmethod
    def install_r_packages(packages: List[str], r_path: Optional[str] = None) -> bool:
        """Install R packages using pak."""
        r_command = r_path if r_path else "R"
        
        console.print(f"[cyan]Installing R packages: {', '.join(packages)}[/cyan]")
        
        try:
            # Create R script for package installation using pak (faster and more reliable)
            packages_str = '", "'.join(packages)
            r_script = f'''
            # Install pak if not available
            if (!require("pak", quietly = TRUE)) {{
                install.packages("pak", repos = "https://cloud.r-project.org/")
            }}
            
            # Install packages using pak
            pak::pkg_install(c("{packages_str}"))
            '''
            
            console.print("[cyan]Installing packages (this may take several minutes)...[/cyan]")
            result = subprocess.run(
                [r_command, "--slave", "--vanilla", "-e", r_script],
                capture_output=True, text=True, timeout=1800  # 30 minutes timeout
            )
            
            if result.returncode == 0:
                console.print("[green]Successfully installed R packages![/green]")
                return True
            else:
                console.print(f"[red]Failed to install R packages: {result.stderr}[/red]")
                return False
                
        except subprocess.TimeoutExpired:
            console.print("[red]Installation timed out. Please install packages manually.[/red]")
            return False
        except Exception as e:
            console.print(f"[red]Error during installation: {e}[/red]")
            return False
    
    @classmethod
    def check_r_packages(cls, r_path: Optional[str] = None) -> List[RPackageCheck]:
        """Check all required R packages."""
        return [cls.check_r_package(pkg, r_path) for pkg in cls.REQUIRED_R_PACKAGES]
    
    @classmethod
    def validate_r_packages(cls, r_path: Optional[str] = None, interactive: bool = True) -> bool:
        """Validate R packages and optionally install missing ones."""
        console.print("[cyan]Checking R packages...[/cyan]")
        
        checks = cls.check_r_packages(r_path)
        missing_packages = [check.name for check in checks if not check.available]
        
        # Display R package status
        cls.display_r_package_status(checks)
        
        if missing_packages:
            if interactive:
                console.print(f"\n[yellow]Found {len(missing_packages)} missing R packages.[/yellow]")
                
                install = Confirm.ask(
                    "Would you like to install the missing R packages automatically?",
                    default=True
                )
                
                if install:
                    success = cls.install_r_packages(missing_packages, r_path)
                    if success:
                        # Re-check packages after installation
                        console.print("\n[cyan]Re-checking R packages...[/cyan]")
                        new_checks = cls.check_r_packages(r_path)
                        still_missing = [check.name for check in new_checks if not check.available]
                        
                        if still_missing:
                            console.print(f"[red]Some packages failed to install: {', '.join(still_missing)}[/red]")
                            cls.display_r_installation_instructions(still_missing)
                            return False
                        else:
                            console.print("[green]All R packages are now available![/green]")
                            return True
                    else:
                        cls.display_r_installation_instructions(missing_packages)
                        return False
                else:
                    cls.display_r_installation_instructions(missing_packages)
                    return False
            else:
                cls.display_r_installation_instructions(missing_packages)
                return False
        else:
            console.print("[green]All required R packages are available![/green]")
            return True
    
    @staticmethod
    def display_r_package_status(checks: List[RPackageCheck]):
        """Display R package status in a formatted table."""
        table = Table(title="R Package Status", show_header=True, header_style="bold cyan")
        table.add_column("Package", style="bold")
        table.add_column("Status", justify="center")
        table.add_column("Version")
        
        for check in checks:
            status = "[green]✓ Available[/green]" if check.available else "[red]✗ Missing[/red]"
            version = check.version or "[dim]Unknown[/dim]"
            table.add_row(check.name, status, version)
        
        console.print(table)
    
    @staticmethod
    def display_r_installation_instructions(missing_packages: List[str]):
        """Display installation instructions for missing R packages."""
        console.print("\n[bold red]Missing R Packages[/bold red]")
        console.print("The following required R packages are missing:\n")
        
        packages_str = '", "'.join(missing_packages)
        
        console.print(f"[red]Missing packages: {', '.join(missing_packages)}[/red]\n")
        
        console.print("[bold]Manual Installation Instructions:[/bold]")
        console.print("You can install these packages manually in R:")
        console.print()
        console.print(f'[dim]# Install pak first (if not available)[/dim]')
        console.print(f'[cyan]install.packages("pak")[/cyan]')
        console.print()
        console.print(f'[dim]# Install all required packages[/dim]')
        console.print(f'[cyan]pak::pkg_install(c("{packages_str}"))[/cyan]')
        console.print()
        console.print("[yellow]Please install the missing R packages before proceeding.[/yellow]")
    
    @classmethod
    def check_all_dependencies(cls) -> List[DependencyCheck]:
        """Check all required and optional dependencies."""
        results = []
        
        for dep in cls.REQUIRED_DEPENDENCIES:
            results.append(cls.check_command(dep))
        
        for dep in cls.OPTIONAL_DEPENDENCIES:
            results.append(cls.check_command(dep))
        
        return results
    
    @classmethod
    def check_required_dependencies(cls) -> List[DependencyCheck]:
        """Check only required dependencies."""
        return [cls.check_command(dep) for dep in cls.REQUIRED_DEPENDENCIES]
    
    @classmethod
    def validate_dependencies(cls, show_details: bool = True, check_r_packages: bool = True, r_path: Optional[str] = None) -> bool:
        """Validate all required dependencies and show results."""
        checks = cls.check_required_dependencies()
        missing_deps = [check for check in checks if not check.available]
        
        if show_details:
            cls.display_dependency_status(checks)
        
        if missing_deps:
            if show_details:
                cls.display_installation_instructions(missing_deps)
        
        # Check R packages even if other dependencies are missing, but only if R is available
        r_available = any(check.name == "R" and check.available for check in checks)
        if check_r_packages and r_available and r_path:
            console.print("\n" + "="*50)
            r_packages_valid = cls.validate_r_packages(r_path, interactive=True)
            if not r_packages_valid:
                return False
        
        # Return False if any system dependencies are missing
        if missing_deps:
            return False
        
        return True
    
    @staticmethod
    def display_dependency_status(checks: List[DependencyCheck]):
        """Display dependency status in a formatted table."""
        table = Table(title="Dependency Status", show_header=True, header_style="bold cyan")
        table.add_column("Dependency", style="bold")
        table.add_column("Status", justify="center")
        table.add_column("Path")
        table.add_column("Version")
        
        for check in checks:
            status = "[green]✓ Available[/green]" if check.available else "[red]✗ Missing[/red]"
            path = check.path or "[dim]Not found[/dim]"
            version = check.version or "[dim]Unknown[/dim]"
            table.add_row(check.name, status, path, version)
        
        console.print(table)
    
    @staticmethod
    def display_installation_instructions(missing_deps: List[DependencyCheck]):
        """Display installation instructions for missing dependencies."""
        console.print("\n[bold red]Missing Dependencies[/bold red]")
        console.print("The following required dependencies are missing:\n")
        
        for dep in missing_deps:
            console.print(f"[red]✗ {dep.name}[/red]")
            
            if dep.name == "cellranger":
                console.print("  Installation: Download from 10x Genomics website")
                console.print("  URL: https://support.10xgenomics.com/single-cell-gene-expression/software/downloads/latest")
                console.print("  After installation, add cellranger to your PATH\n")
                
            elif dep.name == "R":
                console.print("  Installation:")
                console.print("  - macOS: brew install r")
                console.print("  - Ubuntu/Debian: sudo apt-get install r-base")
                console.print("  - CentOS/RHEL: sudo yum install R")
                console.print("  - Or download from: https://cran.r-project.org/\n")
                
                
            elif dep.name == "fastq-dump":
                console.print("  Installation: Install SRA Toolkit")
                console.print("  - macOS: brew install sratoolkit")
                console.print("  - Ubuntu/Debian: sudo apt-get install sra-toolkit")
                console.print("  - Or download from: https://github.com/ncbi/sra-tools\n")
        
        console.print("[yellow]Please install the missing dependencies before proceeding.[/yellow]")
    
    @staticmethod
    def get_rig_installations() -> List[tuple]:
        """Get R installations managed by rig."""
        installations = []
        
        try:
            result = subprocess.run(
                ["rig", "list"], 
                capture_output=True, text=True, timeout=10
            )
            if result.returncode == 0:
                lines = result.stdout.strip().split('\n')
                for line in lines:
                    if line.strip() and not line.startswith('*'):
                        # Parse rig list output: "4.3.2 /usr/local/lib/R/4.3.2"
                        parts = line.strip().split()
                        if len(parts) >= 2:
                            version = parts[0]
                            path = parts[1]
                            r_executable = os.path.join(path, "bin", "R")
                            if os.path.exists(r_executable):
                                installations.append((version, r_executable))
        except (subprocess.SubprocessError, subprocess.TimeoutExpired, FileNotFoundError):
            pass
        
        return installations
    
    @staticmethod
    def create_rig_environment(r_version: str = "release") -> Optional[str]:
        """Create a new R environment using rig."""
        console.print(f"[cyan]Creating R environment with rig (version: {r_version})...[/cyan]")
        
        try:
            # Install the R version if not already installed
            console.print(f"[cyan]Installing R {r_version}...[/cyan]")
            
            # Use rig add without sudo - rig handles user installations
            cmd = ["rig", "add", r_version]
            console.print(f"[dim]Running: {' '.join(cmd)}[/dim]")
            
            # Run with real-time output
            process = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True,
                bufsize=1,
                universal_newlines=True
            )
            
            # Show real-time output
            output_lines = []
            while True:
                output = process.stdout.readline()
                if output == '' and process.poll() is not None:
                    break
                if output:
                    output_lines.append(output.strip())
                    console.print(f"[dim]{output.strip()}[/dim]")
            
            # Wait for completion
            return_code = process.poll()
            
            if return_code != 0:
                console.print(f"[red]Failed to install R {r_version}[/red]")
                console.print(f"[red]Exit code: {return_code}[/red]")
                if output_lines:
                    console.print("[red]Output:[/red]")
                    for line in output_lines[-10:]:  # Show last 10 lines
                        console.print(f"[red]{line}[/red]")
                return None
            
            console.print(f"[green]Successfully installed R {r_version}[/green]")
            
            # Get the path of the installed R
            list_result = subprocess.run(
                ["rig", "list"], 
                capture_output=True, text=True, timeout=10
            )
            
            if list_result.returncode == 0:
                lines = list_result.stdout.strip().split('\n')
                for line in lines:
                    if line.strip():
                        parts = line.strip().split()
                        if len(parts) >= 2:
                            version_part = parts[0]
                            path = parts[1]
                            # Check if this matches our requested version
                            if (r_version == "release" and "*" in line) or r_version in version_part:
                                r_executable = os.path.join(path, "bin", "R")
                                if os.path.exists(r_executable):
                                    console.print(f"[green]R {version_part} is available at: {r_executable}[/green]")
                                    return r_executable
            
            console.print(f"[red]Could not find installed R {r_version}[/red]")
            return None
            
        except subprocess.TimeoutExpired:
            console.print("[red]R installation timed out[/red]")
            return None
        except Exception as e:
            console.print(f"[red]Error creating R environment: {e}[/red]")
            return None
    
    @staticmethod
    def install_r_packages_with_rig(r_path: str, packages: List[str]) -> bool:
        """Install R packages using the rig-managed R installation."""
        console.print(f"[cyan]Installing R packages with rig-managed R: {', '.join(packages)}[/cyan]")
        
        try:
            # Create R script for package installation
            packages_str = '", "'.join(packages)
            r_script = f'''
            # Install pacman if not available
            if (!require("pak", quietly = TRUE)) {{
                install.packages("pak", repos = "https://cloud.r-project.org/")
            }}
            
            # Install packages using pak (faster than install.packages)
            pak::pkg_install(c("{packages_str}"))
            '''
            
            console.print("[cyan]Installing packages (this may take several minutes)...[/cyan]")
            result = subprocess.run(
                [r_path, "--slave", "--vanilla", "-e", r_script],
                capture_output=True, text=True, timeout=1800  # 30 minutes timeout
            )
            
            if result.returncode == 0:
                console.print("[green]Successfully installed R packages![/green]")
                return True
            else:
                console.print(f"[red]Failed to install R packages: {result.stderr}[/red]")
                return False
                
        except subprocess.TimeoutExpired:
            console.print("[red]Package installation timed out[/red]")
            return False
        except Exception as e:
            console.print(f"[red]Error during package installation: {e}[/red]")
            return False
    
    @staticmethod
    def select_r_installation() -> Optional[str]:
        """Interactive R installation selection."""
        console.print("\n[cyan]Selecting R installation...[/cyan]")
        
        # Get current R path
        current_r = shutil.which("R")
        
        # Create choices for inquirer
        choices = []
        
        if current_r:
            # Get version info for current R
            try:
                result = subprocess.run([current_r, "--version"], capture_output=True, text=True, timeout=10)
                if result.returncode == 0:
                    lines = result.stdout.strip().split('\n')
                    version = "Unknown"
                    for line in lines:
                        if "R version" in line:
                            version = line.split()[2]
                            break
                    choices.append(f"Use current R installation (R {version}) - {current_r}")
                else:
                    choices.append(f"Use current R installation - {current_r}")
            except:
                choices.append(f"Use current R installation - {current_r}")
        
        choices.append("Enter custom R path manually")
        
        # Use inquirer for selection
        questions = [
            inquirer.List(
                'r_selection',
                message="Select R installation to use",
                choices=choices,
                carousel=True
            )
        ]
        
        try:
            result = inquirer.prompt(questions, raise_keyboard_interrupt=True)
            if result is None:
                return None
            
            selection = result['r_selection']
            
            if selection == "Enter custom R path manually":
                # Manual path input
                manual_questions = [
                    inquirer.Path(
                        'r_path',
                        message="Enter the full path to R executable",
                        path_type=inquirer.Path.FILE
                    )
                ]
                manual_result = inquirer.prompt(manual_questions, raise_keyboard_interrupt=True)
                if manual_result is None:
                    return None
                
                manual_path = manual_result['r_path']
                
                # Validate the manual path
                if os.path.exists(manual_path) and os.access(manual_path, os.X_OK):
                    console.print(f"[green]Selected R: {manual_path}[/green]")
                    return manual_path
                else:
                    console.print(f"[red]Invalid R path: {manual_path}[/red]")
                    return None
            else:
                # Extract actual path from current R selection
                if " - " in selection:
                    r_path = selection.split(" - ")[1]
                    console.print(f"[green]Selected R: {r_path}[/green]")
                    return r_path
                else:
                    console.print(f"[red]Could not parse selection: {selection}[/red]")
                    return None
                
        except KeyboardInterrupt:
            console.print("\n[yellow]R selection cancelled.[/yellow]")
            return None


# from dataclasses import dataclass
# import datetime
# import os
# import sys
# from typing import (
#     TYPE_CHECKING,
#     Callable,
#     Dict,
#     Final,
#     List,
#     NamedTuple,
#     Optional,
#     Union,
# )

# import polars as pl
# import rich
# from rich.progress import track
# import toml
# import tqdm

# from celline.DB.dev.handler import HandleResolver
# from celline.DB.dev.model import SampleSchema
# from celline.config import Config, Setting
# from celline.functions._base import CellineFunction
# from celline.middleware import ThreadObservable
# from celline.sample import SampleInfo, SampleResolver
# from celline.server import ServerSystem
# from celline.template import TemplateManager
# from celline.utils.serialization import NamedTupleAndPolarsStructure

# if TYPE_CHECKING:
#     from celline import Project


# class BatchCorrection(CellineFunction):
#     """
#     Correction batch effect
#     """

#     def __init__(
#         self,
#         output_file_path: str,
#         filter_func: Optional[Callable[[SampleSchema], bool]],
#     ) -> None:
#         """
#         #### Add accession ID to DB & your project.

#         #### Note: Parallel calculations are not supported

#         Args:
#             h5ad_path (<List[Add.SampleInfo]> | <pl.DataFrame>): Accession ID to add.
#         """
#         self.filter_func = filter_func
#         self.output_file_path = output_file_path
#         self.cluster_server: Final[Optional[str]] = ServerSystem.cluster_server_name

#     def register(self) -> str:
#         return "batch"

#     def call(self, project: "Project") -> "Project":
#         """
#         Call the function to add accession IDs to the project.

#         Args:
#             project (<Project>): The project to add the accession IDs to.

#         Returns:
#             <Project>: The project with the added accession IDs.
#         """
#         # STEP0: Data collection
#         target_samples: list[SampleInfo] = []
#         for info in SampleResolver.samples.values():
#             if info.path.is_counted:
#                 if self.filter_func is None:
#                     add = True
#                 else:
#                     add = self.filter_func(info.schema)
#                 if add:
#                     target_samples.append(info)
#             else:
#                 rich.print(
#                     f":warning: [bold yellow]Warning[/] Target sample {info.schema.key} is not counted or preprocessed yet."
#                 )
#         os.makedirs(f"{Config.PROJ_ROOT}/batch", exist_ok=True)
#         os.makedirs(f"{Config.PROJ_ROOT}/batch/logs", exist_ok=True)
#         os.makedirs(f"{Config.PROJ_ROOT}/batch/logs/runtime", exist_ok=True)
#         # [OBS: STEP1]: Pure integrate
#         # self.__pure_merge(target_samples)
#         # [OBS: STEP2]: Convert rds to h5ad
#         # self.__convert(project)
#         # STEP3: Correct batch effect
#         self.__correct_batch_effect(target_samples)

#         return project

#     def __correct_batch_effect(self, target_samples: list[SampleInfo]):
#         class JobContainer(NamedTuple):
#             nthread: str
#             cluster_server: str
#             logpath: str
#             r_path: str
#             exec_root: str
#             output_dir: str
#             sample_ids: str
#             project_ids: str
#             logpath_runtime: str
#             proj_path: str

#         NOW: Final[str] = datetime.datetime.now().strftime("%Y%m%d%H%M%S")
#         TemplateManager.replace_from_file(
#             file_name="batch_cor.sh",
#             structure=JobContainer(
#                 nthread=str(1),
#                 cluster_server=""
#                 if self.cluster_server is None
#                 else self.cluster_server,
#                 logpath=f"{Config.PROJ_ROOT}/batch/logs/removebatch_{NOW}.log",
#                 r_path=f"{Setting.r_path}script",
#                 exec_root=Config.EXEC_ROOT,
#                 proj_path=Config.PROJ_ROOT,
#                 sample_ids=",".join(
#                     [str(sample.schema.key) for sample in target_samples]
#                 ),
#                 project_ids=",".join(
#                     [
#                         sample.schema.parent if sample.schema.parent is not None else ""
#                         for sample in target_samples
#                     ]
#                 ),
#                 output_dir=self.output_file_path,
#                 logpath_runtime=f"{Config.PROJ_ROOT}/batch/logs/removebatch_{NOW}.sh",
#             ),
#             replaced_path=f"{Config.PROJ_ROOT}/batch/removebatch_{NOW}.sh",
#         )
#         ThreadObservable.call_shell(
#             [f"{Config.PROJ_ROOT}/batch/removebatch_{NOW}.sh"]
#         ).watch()



from __future__ import annotations

from dataclasses import dataclass

# import celline.DB._base import
from typing import Final
from xml.etree import ElementTree as ET
from xml.etree.ElementTree import Element

import requests
from pysradb.sraweb import SRAweb

from celline.DB.dev.model import BaseModel, BaseSchema

# from celline.utils.type import pl_ptr


@dataclass
class SRA_GSE_Schema(BaseSchema):
    summary: str


class SRA_GSE(BaseModel[SRA_GSE_Schema]):
    DB: Final[SRAweb] = SRAweb()

    @classmethod
    def build_request_url(cls, gsm_id: str):
        return f"https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc={gsm_id}&targ=all&view=quick&form=xml"

    @classmethod
    def __fetch_result(cls, gsm_id: str):
        def _strip_namespace(elem: Element) -> None:
            """
            Remove the namespace from the XML tags.
            """
            elem.tag = elem.tag.split("}", 1)[-1] if "}" in elem.tag else elem.tag
            for child in elem:
                _strip_namespace(child)

        try:
            xml = requests.get(
                cls.build_request_url(gsm_id),
                timeout=100,
            )
        except ET.ParseError as err:
            print(f"Could not find target GSE: {gsm_id}")
            print("Tracebacks")
            print(err)
        gse_xml = ET.fromstring(xml.content.decode())  # type: ignore
        _strip_namespace(gse_xml)
        return gse_xml

    def set_class_name(self) -> str:
        return __class__.__name__

    def def_schema(self) -> type[SRA_GSE_Schema]:
        return SRA_GSE_Schema

    def search(self, acceptable_id: str, force_search=False) -> SRA_GSE_Schema:
        cache = self.get_cache(acceptable_id, force_search)
        if cache is not None:
            return cache
        __result = SRA_GSE.__fetch_result(acceptable_id)
        ser = __result.find("Series")
        if ser is None:
            raise ModuleNotFoundError(
                f"Could not found Series element: {acceptable_id}"
            )
        title = ser.find("Title")
        if title is None:
            raise ModuleNotFoundError(f"Could not found Title element: {acceptable_id}")
        summary = ser.find("Summary")
        if summary is None:
            raise ModuleNotFoundError(
                f"Could not found Summary element: {acceptable_id}"
            )
        return self.add_schema(
            SRA_GSE_Schema(
                key=acceptable_id,
                title=title.text if title.text is not None else "Unknown",
                summary=summary.text if summary.text is not None else "Unknown",
                children=",".join([
                    el.attrib["iid"] for el in __result.findall("Sample")
                ]),
                parent=None,
            )
        )



import os
from pathlib import Path
from typing import List, Optional

import numpy as np
import pandas as pd
import polars as pl
import pyper as pr
import scanpy as sc

from celline.config import Config, Setting
from celline.data.ggplot import ggplot
from celline.utils.r_wrap import as_r_bool, as_r_NULL, as_r_nullablestr


class Seurat:
    r: pr.R

    def __init__(self, seurat_path: str, via_seurat_disk: bool) -> None:
        if not os.path.isfile(seurat_path):
            raise FileNotFoundError(
                f"Could not find seurat file: {seurat_path}. Please try consider identifier (file name), or using seurat_from_rawpath"
            )
        print(f"Using R: {Setting.r_path}")
        self.r = pr.R(RCMD=Setting.r_path, use_pandas=True)
        self.r.assign("h5seurat_path", f"{seurat_path}")
        self.r("pacman::p_load(Seurat, SeuratDisk, tidyverse)")
        print("Loading seurat")
        if via_seurat_disk:
            result = self.r("seurat <- SeuratDisk::LoadH5Seurat(h5seurat_path)")
        else:
            result = self.r("seurat <- readRDS(h5seurat_path)")
        print(
            f"""
Done.
---- Trace --------------
{result}
"""
        )

    @property
    def metadata(self) -> pl.DataFrame:
        _metadata = pd.DataFrame(
            self.r.get('seurat@meta.data %>% tibble::rownames_to_column("cell")')
        )
        for col in _metadata.columns:
            if _metadata[col].apply(lambda x: isinstance(x, bytes)).any():
                _metadata[col] = _metadata[col].str.decode("utf-8")
        return pl.from_pandas(_metadata.convert_dtypes())

    def DimPlot(
        self,
        group_by: Optional[str] = "seurat_clusters",
        split_by: Optional[str] = None,
        pt_size: Optional[int] = None,
    ):
        if group_by is None:
            group_by = "NULL"
        if split_by is None:
            split_by = "NULL"

        log = self.r(
            f'plt <- DimPlot(seurat, group.by = "{group_by}", split.by = {as_r_nullablestr(split_by)}, pt.size = {as_r_NULL(pt_size)})'
        )
        print(log)
        return ggplot(self.r)

    def DotPlot(
        self,
        features: List[str],
        assay: Optional[str] = None,
        cols: Optional[List[str]] = None,
        col_min: float = -2.5,
        col_max: float = 2.5,
        dot_min=0,
        dot_scale=6,
        idents: Optional[str] = None,
        group_by: Optional[str] = None,
        split_by: Optional[str] = None,
        cluster_idents=False,
        scale=True,
        scale_by="radius",
        scale_min: Optional[float] = None,
        scale_max: Optional[float] = None,
    ):
        if features is None:
            features = []
        if cols is None:
            cols = ["lightgrey", "blue"]
        self.r.assign("features", features)
        self.r.assign("assay", assay)
        self.r.assign("cols", cols)
        self.r.assign("col.min", col_min)
        self.r.assign("col.max", col_max)
        self.r.assign("dot.min", dot_min)
        self.r.assign("dot.scale", dot_scale)
        self.r.assign("group.by", group_by)
        self.r.assign("idents", idents)
        self.r.assign("split.by", split_by)
        self.r.assign("cluster.idents", cluster_idents)
        self.r.assign("scale", scale)
        self.r.assign("scale.by", scale_by)
        self.r.assign("scale.min", scale_min)
        self.r.assign("scale.max", scale_max)
        cmd = """
plt <-
    seurat %>%
    Seurat::DotPlot(
        assay = assay,
        features,
        cols = cols),
        col.min = col.min,
        col.max = col.max,
        dot.min = dot.min,
        dot.scale = dot.scale,
        idents = idents,
        group.by = group.by,
        split.by = split.by,
        cluster.idents = cluster.idents,
        scale = scale,
        scale.by = scale.by,
        scale.min = scale.min,
        scale.max = scale.max
    )
"""
        self.r(cmd)
        return ggplot(self.r)

    def VlnPlot(
        self,
        features: List[str],
        cols: Optional[List[str]] = None,
        pt_size: Optional[float] = None,
        alpha: float = 1,
        idents: Optional[List[str]] = None,
        sort=False,
        assay: Optional[str] = None,
        group_by: Optional[str] = None,
        split_by: Optional[str] = None,
        adjust: float = 1,
        y_max: Optional[float] = None,
        same_y_lims=False,
        log=False,
        ncol: Optional[int] = None,
        layer="data",
        split_plot=False,
        stack=False,
        combine=True,
        fill_by="feature",
        flip=False,
        add_noise=True,
        raster: Optional[bool] = None,
    ):
        self.r.assign("features", features)
        self.r.assign("cols", cols)
        self.r.assign("pt.size", pt_size)
        self.r.assign("alpha", alpha)
        self.r.assign("idents", idents)
        self.r.assign("sort", sort)
        self.r.assign("assay", assay)
        self.r.assign("group.by", group_by)
        self.r.assign("split.by", split_by)
        self.r.assign("adjust", adjust)
        self.r.assign("y.max", y_max)
        self.r.assign("same.y.lims", same_y_lims)
        self.r.assign("log", log)
        self.r.assign("ncol", ncol)
        self.r.assign("layer", layer)
        self.r.assign("split.plot", split_plot)
        self.r.assign("stack", stack)
        self.r.assign("combine", combine)
        self.r.assign("fill.by", fill_by)
        self.r.assign("flip", flip)
        self.r.assign("add.noise", add_noise)
        self.r.assign("raster", raster)
        self.r(
            """
plt <-
    seurat %>%
    Seurat::VlnPlot(
        features,
        cols = cols,
        pt.size = pt.size,
        alpha = alpha,
        idents = idents,
        sort = sort,
        assay = assay,
        group.by = group.by,
        split.by = split.by,
        adjust = adjust,
        y.max = y.max,
        same.y.lims = same.y.lims,
        log = log,
        ncol = ncol,
        layer = layer,
        split.plot = split.plot,
        stack = stack,
        combine = combine,
        fill.by = fill.by,
        flip = flip,
        add.noise = add.noise,
        raster = raster
    )
"""
        )
        return ggplot(self.r)

    def save(self, path: str):
        # self.r.assign("savepath", path)
        cmd = f"""
seurat %>%
    saveRDS({path})
"""
        self.r(cmd)
        return

    def save_h5ad(self, path: str):
        # self.r.assign("savepath", path)
        os.makedirs(f"{Config.PROJ_ROOT}/cache", exist_ok=True)
        cmd = f"""
pacman::p_load(Matrix, tidyverse)
seurat %>%
    LayerData() %>%
    t() %>%
    writeMM("{Config.PROJ_ROOT}/cache/matrix.mtx")
seurat@meta.data %>%
    tibble::rownames_to_column("barcodes") %>%
    write_csv(paste0("{Config.PROJ_ROOT}/cache/barcodes.csv"))
rownames(seurat) %>%
    as_tibble() %>%
    dplyr::rename(features = "value") %>%
    write_csv(paste0("{Config.PROJ_ROOT}/cache/features.csv"))
"""
        self.r(cmd)
        adata = sc.read_mtx(Path(f"{Config.PROJ_ROOT}/cache/matrix.mtx"), dtype="int32")
        adata.obs = pd.read_csv(f"{Config.PROJ_ROOT}/cache/barcodes.csv")
        adata.obs.set_index("barcodes", inplace=True)
        adata.var = pd.read_csv(f"{Config.PROJ_ROOT}/cache/features.csv")
        adata.write_h5ad(Path(path))
        os.remove(f"{Config.PROJ_ROOT}/cache/matrix.mtx")
        return



from pyper import R
from multipledispatch import dispatch
import polars as pl
from typing import Dict, NamedTuple, overload, Optional

from celline.utils.r_wrap import as_r_bool, as_r_NULL
from celline.utils.exceptions import RSessionNotFoundException, RException


class ggplot:
    class aes(NamedTuple):
        x: str
        y: str
        color: Optional[str]
        fill: Optional[str]

    @overload
    def __init__(self, r_session: R) -> None:
        ...

    @overload
    def __init__(self, r_session: R, aes_def: aes) -> None:
        ...

    @overload
    def __init__(self, r_session: R, aes_def: aes, df: pl.DataFrame) -> None:
        ...

    def __init__(
        self,
        r_session: Optional[R] = None,
        aes_def: Optional[aes] = None,
        df: Optional[pl.DataFrame] = None,
    ) -> None:
        if r_session is None:
            raise RSessionNotFoundException("Please give rsession.")
        self.r_session = r_session
        if aes_def is None and df is not None:
            raise RException("You should aes_def and df")
        self.aes_def = aes_def
        self.df = aes_def

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, name: str):
        self._title = name

    _title: Optional[str] = None
    _x_name: Optional[str]
    _y_name: Optional[str]

    def labs(
        self,
        title: Optional[str] = None,
        x: Optional[str] = None,
        y: Optional[str] = None,
    ):
        if title is not None:
            self._title = title
        if x is not None:
            self._x_name = x
        if y is not None:
            self._y_name = y
        return self

    def ggsave(
        self,
        file_path: str,
        width: int,
        height: int,
        units="cm",
        dpi=300,
        limitsize=True,
    ):
        self.r_session.assign(
            "ggplot.title", self.title if self.title is not None else "title"
        )
        log = self.r_session(
            f"""
plt <-
    plt +
    labs(
        title = ggplot.title
    )
ggsave(filename = "{file_path}", plot = plt, width = {width}, height = {height}, units = "{units}", dpi = {dpi}, limitsize = {as_r_bool(limitsize)})
"""
        )
        print(log)



from typing import Union
from celline.DB.dev.handler import BaseHandler
from celline.DB.model import SRA_GSE, SRA_GSM, SRA_SRR

from tqdm import tqdm
from IPython.display import display, clear_output


class GEOHandler(BaseHandler[SRA_GSE, SRA_GSM, SRA_SRR]):
    def resolver(self, acceptable_id: str):
        self._project = SRA_GSE()
        self._sample = SRA_GSM()
        self._run = SRA_SRR()
        if acceptable_id.startswith("GSE"):
            return SRA_GSE
        if acceptable_id.startswith("GSM"):
            return SRA_GSM
        if acceptable_id.startswith("SRR"):
            return SRA_SRR
        return None

    # @staticmethod
    # def sync(target_gsm_id: str, force_search=False, log=True) -> None:
    #     srr_instance = SRA_SRR()
    #     schema = SRA_GSM().search(target_gsm_id, force_search)
    #     ## Read Parent
    #     _ = SRA_GSE().search(schema.parent_gse_id, force_search)
    #     ## Read Child
    #     ids = schema.child_srr_ids.split(",")
    #     cnt = 1
    #     for srr in ids:
    #         if log:
    #             print(f"--> Migrating {srr} ({cnt}/{len(ids)})")
    #         _ = srr_instance.search(srr, force_search)
    #         cnt += 1
    #     # gses = gse_instance.as_schema(GSE.Schema)
    #     # for gse in gses:
    #     #     if log:
    #     #         print(f"Migrating GSE: {gse.accession_id}")
    #     #     for gsm_id in gse.child_gsm_ids.split(","):
    #     #         if log:
    #     #             print(f"--> Migrating GSM: {gsm_id}")
    #     #         for srr_id in gsm_instance.search(gsm_id).child_srr_ids.split(","):
    #     #             if log:
    #     #                 print(f"--> --> Migrating SRR: {srr_id}")
    #     #             srr_instance.search(srr_id)
    #     # for srr in srr_instance.as_schema(SRR.Schema):
    #     #     gse_instance.search(gsm_instance.search(srr.parent_gsm).parent_gse_id)
    #     # del gse_instance, gsm_instance, srr_instance
    #     return

    # # def __init__(self, acceptable_id: str) -> None:
    # #     self._project = SRA_GSE()
    # #     self._sample = SRA_GSM()
    # #     self._run = SRA_SRR()
    # #     super().__init__(acceptable_id)



import os
import subprocess
from typing import TYPE_CHECKING, Callable, Dict, Final, List, NamedTuple, Optional

import rich
from rich.progress import track

from celline.config import Config, Setting
from celline.functions._base import CellineFunction
from celline.middleware import ThreadObservable
from celline.sample import SampleResolver
from celline.server import ServerSystem
from celline.template import TemplateManager

if TYPE_CHECKING:
    from celline import Project


class Reduce(CellineFunction):
    FILES_TO_KEEP: Final = [
        "outs/filtered_feature_bc_matrix.h5",
        "outs/molecule_info.h5",
        "outs/web_summary.html",
        "_log",
        "outs/filtered_feature_bc_matrix/matrix.mtx.gz",
        "outs/filtered_feature_bc_matrix/features.tsv.gz",
        "outs/filtered_feature_bc_matrix/barcodes.tsv.gz",
    ]

    def call(self, project: "Project"):
        for sample in track(
            SampleResolver.samples.values(),
            description="Processing reducing files...",
        ):
            target_path = sample.path.resources_sample_counted
            if os.path.exists(target_path):
                for foldername, subfolders, filenames in os.walk(
                    target_path, topdown=False, followlinks=True
                ):
                    for filename in filenames:
                        rel_path = os.path.relpath(
                            os.path.join(foldername, filename), target_path
                        )
                        if rel_path not in Reduce.FILES_TO_KEEP:
                            os.remove(os.path.join(foldername, filename))

                    # 2. 空のディレクトリやシンボリックリンクを削除
                    for subfolder in subfolders:
                        full_subfolder_path = os.path.join(foldername, subfolder)
                        if os.path.islink(full_subfolder_path):
                            os.remove(full_subfolder_path)
                        elif os.path.isdir(full_subfolder_path) and not os.listdir(
                            full_subfolder_path
                        ):
                            os.rmdir(full_subfolder_path)
        return project



import argparse
import datetime
import os
import shutil
import subprocess
from collections.abc import Callable
from typing import TYPE_CHECKING, List, NamedTuple, Optional

from rich.console import Console

from celline.DB.dev.handler import HandleResolver
from celline.DB.dev.model import RunSchema, SampleSchema
from celline.functions._base import CellineFunction
from celline.middleware import ThreadObservable
from celline.sample import SampleResolver
from celline.server import ServerSystem
from celline.template import TemplateManager
from celline.utils.path import Path

if TYPE_CHECKING:
    from celline import Project

console = Console()


class Download(CellineFunction):
    """#### Download data into your project."""

    class JobContainer(NamedTuple):
        """Represents job information for data download."""

        filetype: str
        nthread: str
        cluster_server: str
        cluster_server_directive: str
        queue_directive: str
        jobname: str
        logpath: str
        sample_id: str
        download_target: str
        download_source: str
        run_ids_str: str

    def __init__(
        self,
        then: Callable[[str], None] | None = None,
        catch: Callable[[subprocess.CalledProcessError], None] | None = None,
    ) -> None:
        """#### Setup download job function with job mode and thread count."""
        self.nthread = 1
        self.then = then if then is not None else lambda _: None
        self.catch = catch if catch is not None else lambda _: None

    def call(self, project: "Project"):
        """Call the Download function to download data into the project."""
        # PBS設定のチェック
        if ServerSystem.job_system == ServerSystem.JobType.PBS:
            if not ServerSystem.is_pbs_configured():
                console.print("[bold red]ERROR: PBS job system selected but not properly configured![/bold red]")
                console.print(f"[yellow]Current status: {ServerSystem.get_status()}[/yellow]")
                console.print("[yellow]Please set cluster server name using: ServerSystem.usePBS('your_cluster_name')[/yellow]")
                console.print("[bold cyan]For testing, automatically setting cluster name to 'test'[/bold cyan]")
                ServerSystem.usePBS('test')
            console.print(f"[green]Using {ServerSystem.get_status()}[/green]")
        all_job_files: list[str] = []
        for sample_id in SampleResolver.samples.keys():
            resolver = HandleResolver.resolve(sample_id)
            if resolver is None:
                raise ReferenceError(f"Could not resolve target sample id: {sample_id}")
            sample_schema: SampleSchema = resolver.sample.search(sample_id)
            if sample_schema.children is None:
                raise NotImplementedError("Children could not found")
            run_schema: RunSchema = resolver.run.search(
                sample_schema.children.split(",")[0],
            )
            filetype = run_schema.strategy
            if sample_schema.parent is None:
                raise ValueError("Sample parent must not be none")
            path = Path(sample_schema.parent, sample_id)
            path.prepare()
            if not path.is_downloaded and not path.is_counted:
                if os.path.exists(path.resources_sample_raw_fastqs):
                    shutil.rmtree(path.resources_sample_raw_fastqs)
                # PBS用のディレクティブを生成
                cluster_server_name = ServerSystem.cluster_server_name or ""
                cluster_server_directive = f":{cluster_server_name}" if cluster_server_name else ""
                queue_directive = cluster_server_name if cluster_server_name else "default"

                job_container = Download.JobContainer(
                    filetype=filetype,
                    nthread=str(self.nthread),
                    cluster_server=cluster_server_name,
                    cluster_server_directive=cluster_server_directive,
                    queue_directive=queue_directive,
                    jobname="Download",
                    logpath=f"{os.path.abspath(path.resources_sample_log)}/download_{datetime.datetime.now().strftime('%Y%m%d_%H:%M:%S')}.log",
                    sample_id=sample_id,
                    download_target=path.resources_sample_raw,
                    download_source=run_schema.raw_link.split(",")[0] if len(run_schema.raw_link.split(",")) > 0 else "",
                    run_ids_str=sample_schema.children,
                )
                
                TemplateManager.replace_from_file(
                    file_name="download.sh",
                    structure=job_container,
                    replaced_path=f"{path.resources_sample_src}/download.sh",
                )
                all_job_files.append(f"{path.resources_sample_src}/download.sh")
        ThreadObservable.call_shell(all_job_files).watch()
        return project

    def add_cli_args(self, parser: argparse.ArgumentParser) -> None:
        """Add CLI arguments for the Download function."""
        parser.add_argument(
            "--nthread",
            "-n",
            type=int,
            default=1,
            help="Number of threads to use for downloading (default: 1)",
        )
        parser.add_argument(
            "--force",
            "-f",
            action="store_true",
            help="Force re-download even if files already exist",
        )

    def cli(self, project: "Project", args: argparse.Namespace | None = None) -> "Project":
        """CLI entry point for Download function."""
        nthread = 1
        if args and hasattr(args, "nthread"):
            nthread = args.nthread

        force = False
        if args and hasattr(args, "force"):
            force = args.force

        console.print(f"[cyan]Starting download with {nthread} thread(s)...[/cyan]")
        if force:
            console.print("[yellow]Force mode: Will re-download existing files[/yellow]")

        # Create Download instance and call it
        download_instance = Download()
        download_instance.nthread = nthread

        # If force mode, we could modify the logic to re-download
        # For now, just call the regular download method
        return download_instance.call(project)

    def get_description(self) -> str:
        """Get description for CLI help."""
        return """Download data files for samples in your project.

This function downloads raw sequencing data files (FASTQ, BAM) for all samples
that have been added to your project but not yet downloaded."""

    def get_usage_examples(self) -> list[str]:
        """Get usage examples for CLI help."""
        return [
            "celline run download",
            "celline run download --nthread 4",
            "celline run download --force",
        ]
